"""
Index object/sigfile loading and signature saving code.

This is the middleware code responsible for loading and saving signatures
in a variety of ways.

---

Command-line functionality goes in sourmash_args.py.

Low-level JSON reading/writing is in signature.py.

Index objects are implemented in the index submodule.

Public API:

* load_file_as_index(filename, ...) -- load a sourmash.Index class
* SaveSignaturesToLocation(filename) - bulk signature output

APIs for plugins to use:

* class Base_SaveSignaturesToLocation - to implement a new output method.

CTB TODO:
* consider replacing ValueError with IndexNotLoaded in the future.
"""

import sys
import os
import gzip
from io import StringIO
import zipfile
import itertools
import traceback

import screed
import sourmash

from . import plugins as sourmash_plugins
from .logging import notify, debug_literal
from .exceptions import IndexNotLoaded

from .index.sqlite_index import load_sqlite_index, SqliteIndex
from .sbtmh import load_sbt_index
from .lca.lca_db import load_single_database
from . import signature as sigmod
from .index import LinearIndex, ZipFileLinearIndex, MultiIndex
from .manifest import CollectionManifest


def load_file_as_index(filename, *, yield_all_files=False):
    """Load 'filename' as a database; generic database loader.

    If 'filename' contains an SBT or LCA indexed database, or a regular
    Zip file, will return the appropriate objects. If a Zip file and
    yield_all_files=True, will try to load all files within zip, not just
    .sig files.

    If 'filename' is a JSON file containing one or more signatures, will
    return an Index object containing those signatures.

    If 'filename' is a directory, will load *.sig underneath
    this directory into an Index object. If yield_all_files=True, will
    attempt to load all files.
    """
    return _load_database(filename, yield_all_files)


def SaveSignaturesToLocation(location):
    """
    Provides a context manager that saves signatures in various output formats.

    Usage:

    with SaveSignaturesToLocation(filename_or_location) as save_sigs:
       save_sigs.add(sig_obj)
    """
    save_list = itertools.chain(_save_classes, sourmash_plugins.get_save_to_functions())
    for priority, cls in sorted(save_list, key=lambda x: x[0]):
        debug_literal(f"trying to match save function {cls}, priority={priority}")

        if cls.matches(location):
            debug_literal(f"{cls} is a match!")
            return cls(location)

    raise Exception(
        f"cannot determine how to open location {location} for saving; this should never happen!?"
    )


### Implementation machinery for _load_databases


def _load_database(filename, traverse_yield_all, *, cache_size=None):
    """Load file as a database - list of signatures, LCA, SBT, etc.

    Return Index object.

    This is an internal function used by other functions in sourmash_args.
    """
    loaded = False

    # load plugins
    plugin_fns = sourmash_plugins.get_load_from_functions()

    # aggregate with default load_from functions & sort by priority
    load_from_functions = sorted(itertools.chain(_loader_functions, plugin_fns))

    # iterate through loader functions, sorted by priority; try them all.
    # Catch ValueError & IndexNotLoaded but nothing else.
    for priority, desc, load_fn in load_from_functions:
        db = None
        try:
            debug_literal(
                f"_load_databases: trying loader fn - priority {priority} - '{desc}'"
            )
            db = load_fn(
                filename, traverse_yield_all=traverse_yield_all, cache_size=cache_size
            )
        except (ValueError, IndexNotLoaded):
            debug_literal(f"_load_databases: FAIL with ValueError: on fn {desc}.")
            debug_literal(traceback.format_exc())
            debug_literal("(continuing past exception)")

        if db is not None:
            loaded = True
            debug_literal("_load_databases: success!")
            break

    if loaded:
        assert db is not None
        return db

    raise ValueError(f"Error while reading signatures from '{filename}'.")


_loader_functions = []


def add_loader(name, priority):
    "decorator to add name/priority to _loader_functions"

    def dec_priority(func):
        _loader_functions.append((priority, name, func))
        return func

    return dec_priority


@add_loader("load from stdin", 10)
def _load_stdin(filename, **kwargs):
    "Load collection from .sig file streamed in via stdin"
    db = None
    if filename == "-":
        # load as LinearIndex, then pass into MultiIndex to generate a
        # manifest.
        lidx = LinearIndex.load(sys.stdin, filename="-")
        db = MultiIndex.load((lidx,), (None,), parent="-")

    return db


@add_loader("load from standalone manifest", 30)
def _load_standalone_manifest(filename, **kwargs):
    from sourmash.index import StandaloneManifestIndex

    try:
        idx = StandaloneManifestIndex.load(filename)
    except gzip.BadGzipFile as exc:
        raise IndexNotLoaded(exc)

    return idx


@add_loader("load from list of paths", 50)
def _multiindex_load_from_pathlist(filename, **kwargs):
    "Load collection from a list of signature/database files"
    db = MultiIndex.load_from_pathlist(filename)

    return db


@add_loader("load from path (file or directory)", 40)
def _multiindex_load_from_path(filename, **kwargs):
    "Load collection from a directory."
    traverse_yield_all = kwargs["traverse_yield_all"]
    db = MultiIndex.load_from_path(filename, traverse_yield_all)

    return db


@add_loader("load SBT", 60)
def _load_sbt(filename, **kwargs):
    "Load collection from an SBT."
    cache_size = kwargs.get("cache_size")

    try:
        db = load_sbt_index(filename, cache_size=cache_size)
    except (FileNotFoundError, TypeError) as exc:
        raise IndexNotLoaded(exc)

    return db


@add_loader("load revindex", 70)
def _load_revindex(filename, **kwargs):
    "Load collection from an LCA database/reverse index."
    db, _, _ = load_single_database(filename)
    return db


@add_loader("load collection from sqlitedb", 20)
def _load_sqlite_db(filename, **kwargs):
    return load_sqlite_index(filename)


@add_loader("load collection from zipfile", 80)
def _load_zipfile(filename, **kwargs):
    "Load collection from a .zip file."
    db = None
    if filename.endswith(".zip"):
        traverse_yield_all = kwargs["traverse_yield_all"]
        try:
            db = ZipFileLinearIndex.load(
                filename, traverse_yield_all=traverse_yield_all
            )
        except FileNotFoundError as exc:
            # turn this into an IndexNotLoaded => proper exception handling by
            # _load_database.
            raise IndexNotLoaded(exc)

    return db


@add_loader("catch FASTA/FASTQ files and error", 1000)
def _error_on_fastaq(filename, **kwargs):
    "This is a tail-end loader that checks for FASTA/FASTQ sequences => err."
    success = False
    try:
        with screed.open(filename) as it:
            _ = next(it)

            success = True
    except:
        pass

    if success:
        raise Exception(
            f"Error while reading signatures from '{filename}' - got sequences instead! Is this a FASTA/FASTQ file?"
        )


### Implementation machinery for SaveSignaturesToLocation


class Base_SaveSignaturesToLocation:
    "Base signature saving class. Track location (if any) and count."

    def __init__(self, location):
        self.location = location
        self.count = 0

    @classmethod
    def matches(cls, location):
        "returns True when this class should handle a specific location"
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __len__(self):
        return self.count

    def open(self):
        pass

    def close(self):
        pass

    def __enter__(self):
        "provide context manager functionality"
        self.open()
        return self

    def __exit__(self, type, value, traceback):
        "provide context manager functionality"
        self.close()

    def add(self, ss):
        self.count += 1

    def add_many(self, sslist):
        for ss in sslist:
            self.add(ss)


def _get_signatures_from_rust(siglist):
    # this function deals with a disconnect between the way Rust
    # and Python handle signatures; Python expects one
    # minhash (and hence one md5sum) per signature, while
    # Rust supports multiple. For now, go through serializing
    # and deserializing the signature! See issue #1167 for more.
    json_str = sigmod.save_signatures_to_json(siglist)
    yield from sigmod.load_signatures_from_json(json_str)


class SaveSignatures_NoOutput(Base_SaveSignaturesToLocation):
    "Do not save signatures."

    def __repr__(self):
        return "SaveSignatures_NoOutput()"

    @classmethod
    def matches(cls, location):
        return location is None

    def open(self):
        pass

    def close(self):
        pass


class SaveSignatures_Directory(Base_SaveSignaturesToLocation):
    "Save signatures within a directory, using md5sum names."

    def __init__(self, location):
        super().__init__(location)

    def __repr__(self):
        return f"SaveSignatures_Directory('{self.location}')"

    @classmethod
    def matches(cls, location):
        "anything ending in /"
        if location:
            return location.endswith("/")

    def close(self):
        pass

    def open(self):
        try:
            os.mkdir(self.location)
        except FileExistsError:
            pass
        except:
            notify(f"ERROR: cannot create signature output directory '{self.location}'")
            sys.exit(-1)

    def add(self, ss):
        super().add(ss)
        md5 = ss.md5sum()

        # don't overwrite even if duplicate md5sum
        outname = os.path.join(self.location, f"{md5}.sig.gz")
        if os.path.exists(outname):
            i = 0
            while 1:
                outname = os.path.join(self.location, f"{md5}_{i}.sig.gz")
                if not os.path.exists(outname):
                    break
                i += 1

        with open(outname, "wb") as fp:
            sigmod.save_signatures_to_json([ss], fp, compression=1)


class SaveSignatures_SqliteIndex(Base_SaveSignaturesToLocation):
    "Save signatures within a directory, using md5sum names."

    def __init__(self, location):
        super().__init__(location)
        self.location = location
        self.idx = None
        self.cursor = None

    @classmethod
    def matches(cls, location):
        "anything ending in .sqldb"
        if location:
            return location.endswith(".sqldb")

    def __repr__(self):
        return f"SaveSignatures_SqliteIndex('{self.location}')"

    def close(self):
        self.idx.commit()
        self.cursor.execute("VACUUM")
        self.idx.close()

    def open(self):
        self.idx = SqliteIndex.create(self.location, append=True)
        self.cursor = self.idx.cursor()

    def add(self, add_sig):
        for ss in _get_signatures_from_rust([add_sig]):
            super().add(ss)
            self.idx.insert(ss, cursor=self.cursor, commit=False)

            # commit every 1000 signatures.
            if self.count % 1000 == 0:
                self.idx.commit()


class SaveSignatures_SigFile(Base_SaveSignaturesToLocation):
    "Save signatures to a .sig JSON file."

    def __init__(self, location):
        super().__init__(location)
        self.keep = []
        self.compress = 0
        if self.location.endswith(".gz"):
            self.compress = 1

    @classmethod
    def matches(cls, location):
        # match anything that is not None or ""
        return bool(location)

    def __repr__(self):
        return f"SaveSignatures_SigFile('{self.location}')"

    def open(self):
        pass

    def close(self):
        if self.location == "-":
            sigmod.save_signatures_to_json(self.keep, sys.stdout)
        else:
            # text mode? encode in utf-8
            mode = "w"
            encoding = "utf-8"

            # compressed? bytes & binary.
            if self.compress:
                encoding = None
                mode = "wb"

            with open(self.location, mode, encoding=encoding) as fp:
                sigmod.save_signatures_to_json(self.keep, fp, compression=self.compress)

    def add(self, ss):
        super().add(ss)
        self.keep.append(ss)


class SaveSignatures_ZipFile(Base_SaveSignaturesToLocation):
    "Save compressed signatures in an uncompressed Zip file."

    def __init__(self, location):
        super().__init__(location)
        self.storage = None

    @classmethod
    def matches(cls, location):
        "anything ending in .zip"
        if location:
            return location.endswith(".zip")

    def __repr__(self):
        return f"SaveSignatures_ZipFile('{self.location}')"

    def close(self):
        # finish constructing manifest object & save
        manifest = CollectionManifest(self.manifest_rows)
        manifest_name = "SOURMASH-MANIFEST.csv"

        manifest_fp = StringIO()
        manifest.write_to_csv(manifest_fp, write_header=True)
        manifest_data = manifest_fp.getvalue().encode("utf-8")

        self.storage.save(manifest_name, manifest_data, overwrite=True, compress=True)

        self.storage.flush()
        self.storage.close()

    def open(self):
        from .sbt_storage import ZipStorage

        do_create = True
        if os.path.exists(self.location):
            do_create = False

        storage = None
        try:
            storage = ZipStorage(self.location, mode="w")
        except zipfile.BadZipFile:
            pass

        if storage is None:
            raise ValueError(f"File '{self.location}' cannot be opened as a zip file.")

        if not storage.subdir:
            storage.subdir = "signatures"

        # now, try to load manifest
        try:
            manifest_data = storage.load("SOURMASH-MANIFEST.csv")
        except (FileNotFoundError, KeyError):
            # if file already exists must have manifest...
            if not do_create:
                raise ValueError(
                    f"Cannot add to existing zipfile '{self.location}' without a manifest"
                )
            self.manifest_rows = []
        else:
            # success! decode manifest_data, create manifest rows => append.
            manifest_data = manifest_data.decode("utf-8")
            manifest_fp = StringIO(manifest_data)
            manifest = CollectionManifest.load_from_csv(manifest_fp)
            self.manifest_rows = list(manifest._select())

        self.storage = storage

    def _exists(self, name):
        try:
            self.storage.load(name)
            return True
        except KeyError:
            return False

    def add(self, add_sig):
        if not self.storage:
            raise ValueError("this output is not open")

        for ss in _get_signatures_from_rust([add_sig]):
            buf = sigmod.save_signatures_to_json([ss], compression=1)
            md5 = ss.md5sum()

            storage = self.storage
            path = f"{storage.subdir}/{md5}.sig.gz"
            location = storage.save(path, buf)

            # update manifest
            row = CollectionManifest.make_manifest_row(
                ss, location, include_signature=False
            )
            self.manifest_rows.append(row)
            super().add(ss)


_save_classes = [
    (10, SaveSignatures_NoOutput),
    (20, SaveSignatures_Directory),
    (30, SaveSignatures_ZipFile),
    (40, SaveSignatures_SqliteIndex),
    (1000, SaveSignatures_SigFile),
]



import os
import sys

from hypothesis import settings, Verbosity
import pytest

import matplotlib.pyplot as plt

plt.rcParams.update({"figure.max_open_warning": 0})

from sourmash_tst_utils import TempDirectory, RunnerContext

sys.stdout = sys.stderr


@pytest.fixture
def runtmp():
    with TempDirectory() as location:
        yield RunnerContext(location)


@pytest.fixture
def run():
    yield RunnerContext(os.getcwd())


@pytest.fixture(params=[True, False])
def track_abundance(request):
    return request.param


@pytest.fixture(params=[True, False])
def dayhoff(request):
    return request.param


@pytest.fixture(params=[True, False])
def hp(request):
    return request.param


@pytest.fixture(params=[True, False])
def keep_identifiers(request):
    return request.param


@pytest.fixture(params=[True, False])
def keep_versions(request):
    return request.param


@pytest.fixture(params=[2, 5, 10])
def n_children(request):
    return request.param


@pytest.fixture(params=["--linear", "--no-linear"])
def linear_gather(request):
    return request.param


@pytest.fixture(params=["--prefetch", "--no-prefetch"])
def prefetch_gather(request):
    return request.param


@pytest.fixture(params=[True, False])
def use_manifest(request):
    return request.param


@pytest.fixture(params=["json", "sql"])
def lca_db_format(request):
    return request.param


@pytest.fixture(params=["csv", "sql"])
def manifest_db_format(request):
    return request.param


@pytest.fixture(params=["sig", "sig.gz", "zip", ".d/", ".sqldb"])
def sig_save_extension(request):
    return request.param


@pytest.fixture(params=["sig", "sig.gz", "zip", ".d/"])
def sig_save_extension_abund(request):
    return request.param


# these should both always succeed for 'sig check' and 'sig collect' output
# manifests.
@pytest.fixture(params=["--abspath", "--relpath"])
def abspath_or_relpath(request):
    return request.param


# this will fail if subdirs used; see #3008. but this ensures v4 behavior of
# sig collect/sig check works, where manifest paths are interpreted relative
# to cwd.
@pytest.fixture(params=["--no-abspath", "--abspath", "--relpath"])
def abspath_relpath_v4(request):
    return request.param


# --- BEGIN - Only run tests using a particular fixture --- #
# Cribbed from: http://pythontesting.net/framework/pytest/pytest-run-tests-using-particular-fixture/
def pytest_collection_modifyitems(items, config):
    fixture_name = config.option.usesfixture
    if fixture_name is not None:
        selected_items = []
        deselected_items = []

        for item in items:
            if fixture_name in getattr(item, "fixturenames", ()):
                selected_items.append(item)
            else:
                deselected_items.append(item)
        config.hook.pytest_deselected(items=deselected_items)
        items[:] = selected_items


# --- END - Only run tests using a particular fixture --- #


def pytest_addoption(parser):
    parser.addoption(
        "--usesfixture",
        action="store",
        default=None,
        help="just run tests that use a particular fixture",
    )

    parser.addoption(
        "--run-hypothesis", action="store_true", help="run hypothesis tests"
    )


def pytest_runtest_setup(item):
    if item.config.getoption("--run-hypothesis"):
        if not any(mark for mark in item.iter_markers(name="hypothesis")):
            pytest.skip("--run-hypothesis option set, running only hypothesis tests")


settings.register_profile("ci", max_examples=1000)
settings.register_profile("dev", max_examples=10)
settings.register_profile("debug", max_examples=10, verbosity=Verbosity.verbose)
settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "default"))



"""
Tests for distance utils.
"""

import pytest
import numpy as np
from sourmash.distance_utils import (
    containment_to_distance,
    get_exp_probability_nothing_common,
    handle_seqlen_nkmers,
    jaccard_to_distance,
    ANIResult,
    ciANIResult,
    jaccardANIResult,
    var_n_mutated,
    set_size_chernoff,
    set_size_exact_prob,
)


def test_aniresult():
    res = ANIResult(0.4, 0.1)
    assert res.dist == 0.4
    assert res.ani == 0.6
    assert res.p_nothing_in_common == 0.1
    assert res.p_exceeds_threshold == True
    # check that they're equivalent
    res2 = ANIResult(0.4, 0.1)
    assert res == res2
    res3 = ANIResult(0.5, 0)
    assert res != res3
    assert res3.p_exceeds_threshold == False


def test_aniresult_bad_distance():
    """
    Fail if distance is not between 0 and 1.
    """
    with pytest.raises(Exception) as exc:
        ANIResult(1.1, 0.1)
    print("\n", str(exc.value))
    assert "distance value 1.1000 is not between 0 and 1!" in str(exc.value)
    with pytest.raises(Exception) as exc:
        ANIResult(-0.1, 0.1)
    print("\n", str(exc.value))
    assert "distance value -0.1000 is not between 0 and 1!" in str(exc.value)


def test_jaccard_aniresult():
    res = jaccardANIResult(0.4, 0.1, jaccard_error=0.03)
    assert res.dist == 0.4
    assert res.ani is None
    assert res.p_nothing_in_common == 0.1
    assert res.jaccard_error == 0.03
    assert res.p_exceeds_threshold == True
    assert res.je_exceeds_threshold == True
    res3 = jaccardANIResult(0.4, 0.1, jaccard_error=0.03, je_threshold=0.1)
    assert res3.je_exceeds_threshold == False
    assert res3.ani == 0.6


def test_jaccard_aniresult_nojaccarderror():
    # jaccard error is None
    with pytest.raises(Exception) as exc:
        jaccardANIResult(0.4, 0.1, None)
    print("\n", str(exc.value))
    assert "Error: jaccard_error cannot be None." in str(exc.value)


def test_ci_aniresult():
    res = ciANIResult(0.4, 0.1, dist_low=0.3, dist_high=0.5)
    print(res)
    assert res.dist == 0.4
    assert res.ani == 0.6
    assert res.p_nothing_in_common == 0.1
    assert res.ani_low == 0.5
    assert res.ani_high == 0.7
    res2 = ciANIResult(0.4, 0.1, dist_low=0.3, dist_high=0.5)
    assert res == res2
    res3 = ciANIResult(0.4, 0.2, dist_low=0.3, dist_high=0.5)
    assert res != res3


def test_containment_to_distance_zero():
    contain = 0
    scaled = 1
    nkmers = 10000
    ksize = 21
    res = containment_to_distance(
        contain, ksize, scaled, n_unique_kmers=nkmers, estimate_ci=True
    )
    print(res)
    # check results
    exp_dist, exp_low, exp_high, pnc = 1.0, 1.0, 1.0, 1.0
    exp_id, exp_idlow, exp_idhigh, pnc = 0.0, 0.0, 0.0, 1.0
    assert res.dist == exp_dist
    assert res.dist_low == exp_low
    assert res.dist_high == exp_high
    assert res.p_nothing_in_common == pnc
    assert res.ani == exp_id
    assert res.ani_low == exp_idlow
    assert res.ani_high == exp_idhigh
    # check without returning ci
    res2 = containment_to_distance(contain, ksize, scaled, n_unique_kmers=nkmers)
    print(res2)
    exp_res = ciANIResult(
        dist=1.0,
        dist_low=1.0,
        dist_high=1.0,
        p_nothing_in_common=1.0,
        p_threshold=0.001,
    )
    assert res2 == exp_res


def test_containment_to_distance_one():
    contain = 1
    scaled = 1
    nkmers = 10000
    ksize = 21
    res = containment_to_distance(
        contain, ksize, scaled, n_unique_kmers=nkmers, estimate_ci=True
    )
    print(res)
    exp_dist, exp_low, exp_high, pnc = 0.0, 0.0, 0.0, 0.0
    exp_id, exp_idlow, exp_idhigh, pnc = 1.0, 1.0, 1.0, 0.0
    assert res.dist == exp_dist
    assert res.dist_low == exp_low
    assert res.dist_high == exp_high
    assert res.p_nothing_in_common == pnc
    assert res.ani == exp_id
    assert res.ani_low == exp_idlow
    assert res.ani_high == exp_idhigh

    # check without returning ci
    res = containment_to_distance(contain, ksize, scaled, n_unique_kmers=nkmers)
    assert res.dist == exp_dist
    assert res.ani == exp_id
    assert res.p_nothing_in_common == pnc
    assert res.ani_low == 1.0
    assert res.ani_high == 1.0


def test_containment_to_distance_scaled1():
    contain = 0.5
    scaled = 1
    nkmers = 10000
    ksize = 21
    res = containment_to_distance(
        contain, ksize, scaled, n_unique_kmers=nkmers, estimate_ci=True
    )
    print(res)
    # check results
    assert res.dist == 0.032468221476108394
    assert res.ani == 0.9675317785238916
    assert res.dist_low == 0.028709912966405623
    assert res.ani_high == 0.9712900870335944
    assert res.dist_high == 0.03647860197289783
    assert res.ani_low == 0.9635213980271021
    assert res.p_nothing_in_common == 0.0
    # without returning ci
    res2 = containment_to_distance(contain, ksize, scaled, n_unique_kmers=nkmers)
    assert (res2.dist, res2.ani, res2.p_nothing_in_common) == (
        0.032468221476108394,
        0.9675317785238916,
        0.0,
    )
    assert (res2.dist, res2.ani, res2.p_nothing_in_common) == (
        res.dist,
        res.ani,
        res.p_nothing_in_common,
    )


def test_containment_to_distance_scaled100():
    contain = 0.1
    scaled = 100
    nkmers = 10000
    ksize = 31
    res = containment_to_distance(
        contain, ksize, scaled, n_unique_kmers=nkmers, estimate_ci=True
    )
    print(res)
    # check results
    assert res.dist == 0.07158545548052564
    assert res.dist_low == 0.05320779238601372
    assert res.dist_high == 0.09055547672455365
    assert res.p_nothing_in_common == 4.3171247410658655e-05
    assert res.p_exceeds_threshold == False


def test_containment_to_distance_scaled100_2():
    contain = 0.5
    scaled = 100
    nkmers = 10000
    ksize = 21
    res = containment_to_distance(
        contain, ksize, scaled, n_unique_kmers=nkmers, estimate_ci=True
    )
    print(res)
    # check results
    assert res.dist == 0.032468221476108394
    assert res.dist_low == 0.023712063916639017
    assert res.dist_high == 0.04309960543965866
    assert res.p_exceeds_threshold == False


def test_containment_to_distance_k10():
    contain = 0.5
    scaled = 100
    nkmers = 10000
    ksize = 10
    res = containment_to_distance(
        contain, ksize, scaled, n_unique_kmers=nkmers, estimate_ci=True
    )
    print(res)
    # check results
    assert res.dist == 0.06696700846319259
    assert res.dist_low == 0.04982777541057476
    assert res.dist_high == 0.08745108232411622
    assert res.p_exceeds_threshold == False


def test_containment_to_distance_confidence():
    contain = 0.1
    scaled = 100
    nkmers = 10000
    ksize = 31
    confidence = 0.99
    res = containment_to_distance(
        contain,
        ksize,
        scaled,
        confidence=confidence,
        n_unique_kmers=nkmers,
        estimate_ci=True,
    )
    print(res)
    # check results
    assert res.dist == 0.07158545548052564
    assert res.dist_low == 0.04802880300938562
    assert res.dist_high == 0.09619930040790341
    assert res.p_exceeds_threshold == False
    confidence = 0.90
    res2 = containment_to_distance(
        contain,
        ksize,
        scaled,
        n_unique_kmers=nkmers,
        confidence=confidence,
        estimate_ci=True,
    )
    print(res2)
    # check results
    assert res2.dist == res.dist
    assert round(res2.dist_low, 10) == round(0.05599435479247415, 10)
    assert res2.dist_high == 0.08758718871990222
    assert res.p_exceeds_threshold == False


def test_nkmers_to_bp_containment():
    containment = 0.1
    scaled = 100
    bp_len = 10030
    ksize = 31
    nkmers = handle_seqlen_nkmers(ksize, sequence_len_bp=bp_len)
    print("nkmers_from_bp:", nkmers)
    confidence = 0.99
    kmer_res = containment_to_distance(
        containment,
        ksize,
        scaled,
        confidence=confidence,
        n_unique_kmers=nkmers,
        estimate_ci=True,
    )
    bp_res = containment_to_distance(
        containment,
        ksize,
        scaled,
        confidence=confidence,
        sequence_len_bp=bp_len,
        estimate_ci=True,
    )
    print(f"\nkDIST: {kmer_res}")
    print(f"\nbpDIST:,{bp_res}")
    # check results
    assert kmer_res == bp_res
    assert kmer_res.dist == 0.07158545548052564
    assert kmer_res.dist_low == 0.04802880300938562
    assert kmer_res.dist_high == 0.09619930040790341


def test_jaccard_to_distance_zero():
    jaccard = 0
    scaled = 1
    nkmers = 10000
    ksize = 21
    res = jaccard_to_distance(jaccard, ksize, scaled, n_unique_kmers=nkmers)
    print(res)
    # check results
    assert res.dist == 1.0
    assert res.ani == 0.0
    assert res.p_nothing_in_common == 1.0
    assert res.jaccard_error == 0.0


def test_jaccard_to_distance_one():
    jaccard = 1
    scaled = 1
    nkmers = 10000
    ksize = 21
    res = jaccard_to_distance(jaccard, ksize, scaled, n_unique_kmers=nkmers)
    print(res)
    # check results
    assert res.dist == 0.0
    assert res.ani == 1.0
    assert res.p_nothing_in_common == 0.0
    assert res.jaccard_error == 0.0


def test_jaccard_to_distance_scaled():
    # scaled value doesn't impact point estimate or jaccard error, just p_nothing_in_common
    jaccard = 0.5
    scaled = 1
    nkmers = 10000
    ksize = 21
    res = jaccard_to_distance(jaccard, ksize, scaled, n_unique_kmers=nkmers)
    print(res)
    # check results
    assert round(res.dist, 3) == round(0.019122659390482077, 3)
    assert res.ani is None
    assert res.p_exceeds_threshold == False
    assert res.jaccard_error == 0.00018351337045518042
    assert res.je_exceeds_threshold == True
    scaled = 100
    res2 = jaccard_to_distance(jaccard, ksize, scaled, n_unique_kmers=nkmers)
    print(res2)
    assert res2.dist == res.dist
    assert res2.jaccard_error == res.jaccard_error
    assert res2.p_nothing_in_common != res.p_nothing_in_common
    assert res2.p_exceeds_threshold == False


def test_jaccard_to_distance_k31():
    jaccard = 0.5
    scaled = 100
    nkmers = 10000
    ksize = 31
    res = jaccard_to_distance(jaccard, ksize, scaled, n_unique_kmers=nkmers)
    print(res)
    # check results
    assert res.je_exceeds_threshold == True
    assert res.ani is None
    assert res.p_exceeds_threshold == False
    res2 = jaccard_to_distance(
        jaccard, ksize, scaled, n_unique_kmers=nkmers, err_threshold=0.1
    )
    assert res2.je_exceeds_threshold == False
    assert res2.ani == 0.9870056455892898


def test_jaccard_to_distance_k31_2():
    jaccard = 0.1
    scaled = 100
    nkmers = 10000
    ksize = 31
    res = jaccard_to_distance(jaccard, ksize, scaled, n_unique_kmers=nkmers)
    print(res)
    # check results
    assert res.ani == 0.9464928391768298
    assert res.p_exceeds_threshold == False
    assert res.je_exceeds_threshold == False


def test_nkmers_to_bp_jaccard():
    jaccard = 0.1
    scaled = 100
    bp_len = 10030
    ksize = 31
    nkmers = handle_seqlen_nkmers(ksize, sequence_len_bp=bp_len)
    print("nkmers_from_bp:", nkmers)
    kmer_res = jaccard_to_distance(jaccard, ksize, scaled, n_unique_kmers=nkmers)
    bp_res = jaccard_to_distance(jaccard, ksize, scaled, sequence_len_bp=bp_len)
    print(f"\nkmer_res: {kmer_res}")
    print(f"\nbp_res: {bp_res}")
    # check results
    assert kmer_res == bp_res
    assert kmer_res.dist == 0.0535071608231702
    assert kmer_res.p_exceeds_threshold == False
    assert kmer_res.je_exceeds_threshold == False


def test_exp_prob_nothing_common():
    dist = 0.25
    ksize = 31
    scaled = 10
    bp_len = 1000030
    nkmers = handle_seqlen_nkmers(ksize, sequence_len_bp=bp_len)
    print("nkmers_from_bp:", nkmers)

    nkmers_pnc = get_exp_probability_nothing_common(
        dist, ksize, scaled, n_unique_kmers=nkmers
    )
    print(f"prob nothing in common: {nkmers_pnc}")
    bp_pnc = get_exp_probability_nothing_common(
        dist, ksize, scaled, sequence_len_bp=bp_len
    )
    assert nkmers_pnc == bp_pnc == 7.437016945722123e-07


def test_containment_to_distance_tinytestdata_var0():
    """
    tiny test data to trigger the following:
    WARNING: Cannot estimate ANI confidence intervals from containment. Do your sketches contain enough hashes?
    Error: varN <0.0!
    """
    contain = 0.9
    scaled = 1
    nkmers = 4
    ksize = 31
    res = containment_to_distance(
        contain, ksize, scaled, n_unique_kmers=nkmers, estimate_ci=True
    )
    print(res)
    # check results
    assert res.dist == 0.003392957179023992
    assert res.dist_low is None
    assert res.dist_high is None
    assert res.ani_low is None
    assert res.ani_high is None
    assert res.p_exceeds_threshold == False


def test_var_n_mutated():
    # check 0
    r = 0
    ksize = 31
    nkmers = 200
    var_n_mut = var_n_mutated(nkmers, ksize, r)
    print(f"var_n_mutated: {var_n_mut}")
    assert var_n_mut == 0
    # check var 0.0 valuerror
    r = 10
    ksize = 31
    nkmers = 200
    with pytest.raises(ValueError) as exc:
        var_n_mut = var_n_mutated(nkmers, ksize, r)
    assert "Error: varN <0.0!" in str(exc)
    # check successful
    r = 0.4
    ksize = 31
    nkmers = 200000
    var_n_mut = var_n_mutated(nkmers, ksize, r)
    print(f"var_n_mutated: {var_n_mut}")
    assert var_n_mut == 0.10611425440741508


def test_handle_seqlen_nkmers():
    bp_len = 10030
    ksize = 31
    # convert seqlen to nkmers
    nkmers = handle_seqlen_nkmers(ksize, sequence_len_bp=bp_len)
    assert nkmers == 10000
    # if nkmers is provided, just use that
    nkmers = handle_seqlen_nkmers(ksize, sequence_len_bp=bp_len, n_unique_kmers=bp_len)
    assert nkmers == 10030
    # if neither seqlen or nkmers provided, complain
    with pytest.raises(ValueError) as exc:
        nkmers = handle_seqlen_nkmers(ksize)
    assert (
        "Error: distance estimation requires input of either 'sequence_len_bp' or 'n_unique_kmers'"
    ) in str(exc)


def test_set_size_chernoff():
    eps = 10 ** (-6)
    rel_error = 0.01
    set_size = 1000000
    s = 1 / 0.1  # I'm used to using a scale value between 0 and 1
    value_from_mathematica = 0.928652
    assert (
        np.abs(
            set_size_chernoff(set_size, s, relative_error=rel_error)
            - value_from_mathematica
        )
        < eps
    )

    rel_error = 0.05
    set_size = 10000
    s = 1
    value_from_mathematica = 0.999519
    assert (
        np.abs(
            set_size_chernoff(set_size, s, relative_error=rel_error)
            - value_from_mathematica
        )
        < eps
    )

    rel_error = 0.001
    set_size = 10
    s = 1 / 0.01
    value_from_mathematica = -1
    assert (
        np.abs(
            set_size_chernoff(set_size, s, relative_error=rel_error)
            - value_from_mathematica
        )
        < eps
    )


def test_set_size_exact_prob():
    # values obtained from Mathematica
    # specifically: Probability[Abs[X*s - n]/n <= delta,
    #   X \[Distributed] BinomialDistribution[n, 1/s]] // N
    set_size = 100
    scaled = 2
    relative_error = 0.05
    prob = set_size_exact_prob(set_size, scaled, relative_error=relative_error)
    true_prob = 0.382701
    np.testing.assert_array_almost_equal(true_prob, prob, decimal=3)

    set_size = 200
    scaled = 5
    relative_error = 0.15
    prob = set_size_exact_prob(set_size, scaled, relative_error=relative_error)
    true_prob = 0.749858
    np.testing.assert_array_almost_equal(true_prob, prob, decimal=3)

    set_size = 10
    scaled = 10
    relative_error = 0.10
    prob = set_size_exact_prob(set_size, scaled, relative_error=relative_error)
    true_prob = 0.38742
    np.testing.assert_array_almost_equal(true_prob, prob, decimal=3)

    set_size = 1000
    scaled = 10
    relative_error = 0.10
    prob = set_size_exact_prob(set_size, scaled, relative_error=relative_error)
    true_prob = 0.73182
    np.testing.assert_array_almost_equal(true_prob, prob, decimal=3)



"""remove abundances"""

usage = """

### `sourmash signature flatten` - remove abundance information from signatures

Flatten the specified signature(s), removing abundances and setting
track_abundance to False.

For example,

sourmash signature flatten *.sig -o flattened.sig

will remove all abundances from all of the .sig files in the current
directory.

The `flatten` command accepts the same selectors as `extract`.

"""

from sourmash.cli.utils import add_moltype_args, add_ksize_arg, add_picklist_args


def subparser(subparsers):
    subparser = subparsers.add_parser("flatten", description=__doc__, usage=usage)
    subparser.add_argument("signatures", nargs="*")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        help="output signature to this file (default stdout)",
        default="-",
    )
    subparser.add_argument(
        "--md5",
        default=None,
        help="select signatures whose md5 contains this substring",
    )
    subparser.add_argument(
        "--name",
        default=None,
        help="select signatures whose name contains this substring",
    )
    subparser.add_argument(
        "-f", "--force", action="store_true", help="try to load all files as signatures"
    )
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_picklist_args(subparser)


def main(args):
    import sourmash

    return sourmash.sig.__main__.flatten(args)



"Various utilities used by sourmash tests."

import sys
import os
import tempfile
import shutil
import subprocess
import collections
import pprint
import traceback
from io import StringIO
from pathlib import Path

from importlib import resources
from importlib.metadata import entry_points


SIG_FILES = [
    os.path.join("demo", f)
    for f in (
        "SRR2060939_1.sig",
        "SRR2060939_2.sig",
        "SRR2241509_1.sig",
        "SRR2255622_1.sig",
        "SRR453566_1.sig",
        "SRR453569_1.sig",
        "SRR453570_1.sig",
    )
]


def scriptpath(scriptname="sourmash"):
    """Return the path to the scripts, in both dev and install situations."""
    # note - it doesn't matter what the scriptname is here, as long as
    # it's some script present in this version of sourmash.

    path = os.path.join(os.path.dirname(__file__), "../")
    if os.path.exists(os.path.join(path, scriptname)):
        return path

    path = os.path.join(os.path.dirname(__file__), "../../EGG-INFO/")
    if os.path.exists(os.path.join(path, scriptname)):
        return path

    for path in os.environ["PATH"].split(":"):
        if os.path.exists(os.path.join(path, scriptname)):
            return path


def _runscript(scriptname):
    """Find & run a script with exec (i.e. not via os.system or subprocess)."""
    namespace = {"__name__": "__main__"}
    namespace["sys"] = globals()["sys"]

    try:
        (script,) = entry_points(name=scriptname, group="console_scripts")
        script.load()()
        return 0
    except ValueError:
        pass

    path = scriptpath()

    scriptfile = os.path.join(path, scriptname)
    if os.path.isfile(scriptfile):
        if os.path.isfile(scriptfile):
            exec(  # pylint: disable=exec-used
                compile(Path(scriptfile).read_text(), scriptfile, "exec"), namespace
            )
            return 0

    return -1


ScriptResults = collections.namedtuple("ScriptResults", ["status", "out", "err"])


def runscript(scriptname, args, **kwargs):
    """Run a Python script using exec().

    Run the given Python script, with the given args, in the given directory,
    using 'exec'.  Mimic proper shell functionality with argv, and capture
    stdout and stderr.

    When using :attr:`fail_ok`=False in tests, specify the expected error.
    """
    __tracebackhide__ = True
    sysargs = [scriptname]
    sysargs.extend(args)

    cwd = os.getcwd()
    in_directory = kwargs.get("in_directory", cwd)
    fail_ok = kwargs.get("fail_ok", False)

    try:
        status = -1
        oldargs = sys.argv
        sys.argv = sysargs

        oldin = None
        if "stdin_data" in kwargs:
            oldin, sys.stdin = sys.stdin, StringIO(kwargs["stdin_data"])

        oldout, olderr = sys.stdout, sys.stderr
        sys.stdout = StringIO()
        sys.stdout.name = "StringIO"
        sys.stderr = StringIO()

        os.chdir(in_directory)

        try:
            print("running:", scriptname, "in:", in_directory, file=oldout)
            print("arguments", sysargs, file=oldout)

            status = _runscript(scriptname)
        except SystemExit as err:
            status = err.code
            if status is None:
                status = 0
        except:  # pylint: disable=bare-except
            traceback.print_exc(file=sys.stderr)
            status = -1
    finally:
        sys.argv = oldargs
        out, err = sys.stdout.getvalue(), sys.stderr.getvalue()
        sys.stdout, sys.stderr = oldout, olderr

        if oldin:
            sys.stdin = oldin

        os.chdir(cwd)

    if status != 0 and not fail_ok:
        print(out)
        print(err)
        assert False, (status, out, err)

    return ScriptResults(status, out, err)


def get_test_data(filename):
    filepath = resources.files("sourmash") / "tests" / "test-data" / filename
    if not filepath.exists() or not os.path.isfile(filepath):
        filepath = os.path.join(os.path.dirname(__file__), "test-data", filename)
    return filepath


class TempDirectory:
    def __init__(self):
        self.tempdir = tempfile.mkdtemp(prefix="sourmashtest_")

    def __enter__(self):
        return self.tempdir

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            shutil.rmtree(self.tempdir, ignore_errors=True)
        except OSError:
            pass

        if exc_type:
            return False


class SourmashCommandFailed(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)
        self.message = msg


class RunnerContext:
    """
    I am a RunnerContext object from sourmash_tst_utils.

    I have methods 'run_sourmash' and 'run', which run Python scripts.

    Take a look at my 'location', 'last_command' and 'last_result' attributes!

    You can use the 'output' method to build filenames in my temp directory.
    """

    def __init__(self, location):
        self.location = location
        self.last_command = None
        self.last_result = None

    def run_sourmash(self, *args, **kwargs):
        "Run the sourmash script with the given arguments."
        kwargs["fail_ok"] = True
        if "in_directory" not in kwargs:
            kwargs["in_directory"] = self.location

        cmdlist = ["sourmash"]
        cmdlist.extend(str(x) for x in args)
        self.last_command = " ".join(cmdlist)
        self.last_result = runscript("sourmash", args, **kwargs)

        if self.last_result.status:
            raise SourmashCommandFailed(self.last_result.err)

        return self.last_result

    sourmash = run_sourmash

    def run(self, scriptname, *args, **kwargs):
        "Run a script with the given arguments."
        if "in_directory" not in kwargs:
            kwargs["in_directory"] = self.location
        self.last_command = " ".join(args)
        self.last_result = runscript(scriptname, args, **kwargs)
        return self.last_result

    def output(self, path):
        return os.path.join(self.location, path)

    def __str__(self):
        s = ""
        if self.last_command:
            s += f"Last command run:\n{repr(self.last_command)}\n"
            if self.last_result:
                s += "\nLAST RESULT:\n"
                s += f"- exit code: {self.last_result.status}\n\n"
                if self.last_result.out:
                    s += f"- stdout:\n---\n{self.last_result.out}---\n"
                else:
                    s += "(no stdout)\n\n"
                if self.last_result.err:
                    s += f"- stderr:\n---\n{self.last_result.err}---\n"
                else:
                    s += "(no stderr)\n"

        return s


def in_tempdir(fn):
    def wrapper(*args, **kwargs):
        with TempDirectory() as location:
            ctxt = RunnerContext(location)
            newargs = [ctxt] + list(args)
            return fn(*newargs, **kwargs)

    return wrapper


def in_thisdir(fn):
    def wrapper(*args, **kwargs):
        ctxt = RunnerContext(os.getcwd())
        newargs = [ctxt] + list(args)
        return fn(*newargs, **kwargs)

    return wrapper



"""
Command-line entry point for 'python -m sourmash.tax'
"""

import sys
import csv
import os
from collections import defaultdict, Counter
from dataclasses import asdict, fields
import re

import sourmash
from ..sourmash_args import FileOutputCSV, FileInputCSV, FileOutput
from sourmash.logging import set_quiet, error, notify, print_results

from . import tax_utils
from .tax_utils import (
    MultiLineageDB,
    RankLineageInfo,
    LINLineageInfo,
    AnnotateTaxResult,
    ICTVRankLineageInfo,
)

usage = """
sourmash taxonomy <command> [<args>] - manipulate/work with taxonomy information.
or
sourmash tax <command> [<args>]


** Commands can be:

annotate -g <gather_csv> [<gather_csv> ... ] -t [<taxonomy_csv> ...]      - annotate gather CSVs with taxonomic lineages
genome -g <gather_csv> [<gather_csv> ... ] -t [<taxonomy_csv> ...]        - taxonomic classification of genomes from gather results
metagenome -g <gather_csv> [<gather_csv> ... ] -t [<taxonomy_csv> ...]    - summarize taxonomic information for metagenome gather results

** Use '-h' to get subcommand-specific help, e.g.

sourmash taxonomy metagenome -h
"""

# outfile utils
_output_type_to_ext = {
    "csv_summary": ".summarized.csv",
    "classification": ".classifications.csv",
    "krona": ".krona.tsv",
    "lineage_summary": ".lineage_summary.tsv",
    "annotate": ".with-lineages.csv",
    "human": ".human.txt",
    "lineage_csv": ".lineage.csv",
    "kreport": ".kreport.txt",
    "lingroup": ".lingroup.tsv",
    "bioboxes": ".bioboxes.profile",
}


def make_outfile(base, output_type, *, output_dir=""):
    limit_float_decimals = False
    if base == "-":
        limit_float_decimals = True
        return base, limit_float_decimals

    ext = _output_type_to_ext[output_type]

    fname = base + ext
    if output_dir:
        fname = os.path.join(output_dir, fname)
    notify(f"saving '{output_type}' output to '{fname}'.")
    return fname, limit_float_decimals


##### taxonomy command line functions
def metagenome(args):
    """
    summarize taxonomic information for metagenome gather results
    """
    set_quiet(args.quiet)

    # first, load taxonomic_assignments
    try:
        tax_assign = MultiLineageDB.load(
            args.taxonomy_csv,
            keep_full_identifiers=args.keep_full_identifiers,
            keep_identifier_versions=args.keep_identifier_versions,
            force=args.force,
            lins=args.lins,
            ictv=args.ictv,
        )
        available_ranks = tax_assign.available_ranks
    except ValueError as exc:
        error(f"ERROR: {str(exc)}")
        sys.exit(-1)

    if not tax_assign:
        error(
            f"ERROR: No taxonomic assignments loaded from {','.join(args.taxonomy_csv)}. Exiting."
        )
        sys.exit(-1)

    if args.rank and args.rank not in available_ranks:
        error(
            f"ERROR: No taxonomic information provided for rank {args.rank}: cannot summarize at this rank"
        )
        sys.exit(-1)

    # next, collect and load gather results
    gather_csvs = tax_utils.collect_gather_csvs(
        args.gather_csv, from_file=args.from_file
    )
    try:
        query_gather_results = tax_utils.check_and_load_gather_csvs(
            gather_csvs,
            tax_assign,
            force=args.force,
            fail_on_missing_taxonomy=args.fail_on_missing_taxonomy,
            keep_full_identifiers=args.keep_full_identifiers,
            keep_identifier_versions=args.keep_identifier_versions,
            lins=args.lins,
            ictv=args.ictv,
        )
    except ValueError as exc:
        error(f"ERROR: {str(exc)}")
        sys.exit(-1)

    if not query_gather_results:
        notify("No gather results loaded. Exiting.")
        sys.exit(-1)

    single_query_output_formats = ["kreport", "lingroup", "bioboxes"]
    desired_single_outputs = []
    if len(query_gather_results) > 1:  # working with multiple queries
        desired_single_outputs = [
            x for x in args.output_format if x in single_query_output_formats
        ]
        if desired_single_outputs:
            notify(
                f"WARNING: found results for multiple gather queries. Can only output multi-query result formats: skipping {', '.join(desired_single_outputs)}"
            )
        # remove single query outputs from output format
        args.output_format = [
            x for x in args.output_format if x not in single_query_output_formats
        ]
        if (
            not args.output_format
        ):  # or do we want to insert `human` here so we always report something?
            error("ERROR: No output formats remaining.")
            sys.exit(-1)

    # for each queryResult, actually summarize at rank, reporting any errors that occur.
    for queryResult in query_gather_results:
        try:
            queryResult.build_summarized_result()
        except ValueError as exc:
            error(f"ERROR: {str(exc)}")
            sys.exit(-1)

    # if lingroup file is passed in, read it
    lingroups = None
    if args.lingroup is not None:
        try:
            lingroups = tax_utils.read_lingroups(args.lingroup)
        except ValueError as exc:
            error(f"ERROR: {str(exc)}")
            sys.exit(-1)

    # write summarized output in human-readable format
    if "lineage_summary" in args.output_format:
        lineage_outfile, limit_float = make_outfile(
            args.output_base, "lineage_summary", output_dir=args.output_dir
        )

        ## aggregate by lineage by query
        lineageD, query_names = tax_utils.aggregate_by_lineage_at_rank(
            query_gather_results=query_gather_results, rank=args.rank, by_query=True
        )

        with FileOutputCSV(lineage_outfile) as out_fp:
            tax_utils.write_lineage_sample_frac(query_names, lineageD, out_fp, sep="\t")

    # write summarized --> krona output tsv
    if "krona" in args.output_format:
        krona_results, header = tax_utils.format_for_krona(
            query_gather_results, rank=args.rank
        )

        krona_outfile, limit_float = make_outfile(
            args.output_base, "krona", output_dir=args.output_dir
        )
        with FileOutputCSV(krona_outfile) as out_fp:
            tax_utils.write_krona(header, krona_results, out_fp)

    if "human" in args.output_format:
        summary_outfile, limit_float = make_outfile(
            args.output_base, "human", output_dir=args.output_dir
        )

        with FileOutput(summary_outfile) as out_fp:
            human_display_rank = args.rank or "species"
            if args.lins and not args.rank:
                human_display_rank = query_gather_results[0].ranks[-1]  # lowest rank

            tax_utils.write_human_summary(
                query_gather_results, out_fp, human_display_rank
            )

    # write summarized output csv
    single_query_results = query_gather_results[0]
    if "csv_summary" in args.output_format:
        summary_outfile, limit_float = make_outfile(
            args.output_base, "csv_summary", output_dir=args.output_dir
        )
        with FileOutputCSV(summary_outfile) as out_fp:
            tax_utils.write_summary(
                query_gather_results,
                out_fp,
                limit_float_decimals=limit_float,
                lingroups=lingroups,
            )

    # write summarized --> kreport output tsv
    if "kreport" in args.output_format:
        kreport_outfile, limit_float = make_outfile(
            args.output_base, "kreport", output_dir=args.output_dir
        )

        with FileOutputCSV(kreport_outfile) as out_fp:
            header, kreport_results = single_query_results.make_kreport_results()
            tax_utils.write_output(
                header, kreport_results, out_fp, sep="\t", write_header=False
            )

    # write summarized --> LINgroup output tsv
    if "lingroup" in args.output_format and lingroups is not None:
        lingroupfile, limit_float = make_outfile(
            args.output_base, "lingroup", output_dir=args.output_dir
        )

        with FileOutputCSV(lingroupfile) as out_fp:
            header, lgreport_results = single_query_results.make_lingroup_results(
                LINgroupsD=lingroups
            )
            tax_utils.write_output(
                header, lgreport_results, out_fp, sep="\t", write_header=True
            )

    # write cami bioboxes format
    if "bioboxes" in args.output_format:
        bbfile, limit_float = make_outfile(
            args.output_base, "bioboxes", output_dir=args.output_dir
        )

        with FileOutputCSV(bbfile) as out_fp:
            header_lines, bb_results = single_query_results.make_cami_bioboxes()
            tax_utils.write_bioboxes(header_lines, bb_results, out_fp, sep="\t")


def genome(args):
    """
    taxonomic classification of genomes from gather results
    """
    set_quiet(args.quiet)

    # first, load taxonomic_assignments
    try:
        tax_assign = MultiLineageDB.load(
            args.taxonomy_csv,
            keep_full_identifiers=args.keep_full_identifiers,
            keep_identifier_versions=args.keep_identifier_versions,
            force=args.force,
            lins=args.lins,
            ictv=args.ictv,
        )
        available_ranks = tax_assign.available_ranks

        lg_ranks = None
        all_lgs = None
        if args.lingroup:
            lingroups = tax_utils.read_lingroups(args.lingroup)
            lg_ranks, all_lgs = tax_utils.parse_lingroups(lingroups)

    except ValueError as exc:
        error(f"ERROR: {str(exc)}")
        sys.exit(-1)

    if not tax_assign:
        error(
            f"ERROR: No taxonomic assignments loaded from {','.join(args.taxonomy_csv)}. Exiting."
        )
        sys.exit(-1)

    if args.rank and args.rank not in available_ranks:
        error(
            f"ERROR: No taxonomic information provided for rank {args.rank}: cannot classify at this rank"
        )
        sys.exit(-1)

    # get gather_csvs from args
    gather_csvs = tax_utils.collect_gather_csvs(
        args.gather_csv, from_file=args.from_file
    )

    try:
        query_gather_results = tax_utils.check_and_load_gather_csvs(
            gather_csvs,
            tax_assign,
            force=args.force,
            fail_on_missing_taxonomy=args.fail_on_missing_taxonomy,
            keep_full_identifiers=args.keep_full_identifiers,
            keep_identifier_versions=args.keep_identifier_versions,
            lins=args.lins,
            ictv=args.ictv,
        )

    except ValueError as exc:
        error(f"ERROR: {str(exc)}")
        sys.exit(-1)

    if not query_gather_results:
        notify("No results for classification. Exiting.")
        sys.exit(-1)

    # for each queryResult, summarize at rank and classify according to thresholds, reporting any errors that occur.
    n_total = len(query_gather_results)
    classified_results = []
    found_error = False
    for queryResult in query_gather_results:
        try:
            queryResult.build_classification_result(
                rank=args.rank,
                ani_threshold=args.ani_threshold,
                containment_threshold=args.containment_threshold,
                lingroup_ranks=lg_ranks,
                lingroups=all_lgs,
            )
            classified_results.append(queryResult)

        except ValueError as exc:
            found_error = True
            notify(f"ERROR: {str(exc)}")

    n_classified = len(classified_results)
    if n_classified == 0:
        notify("No queries could be classified. Exiting.")
        sys.exit(-1)
    else:
        classif_perc = (float(n_classified) / float(n_total)) * 100
        notify(
            f"classified {n_classified}/{n_total} queries ({classif_perc:.2f}%). Writing results"
        )

    # write outputs
    if "csv_summary" in args.output_format:
        summary_outfile, limit_float = make_outfile(
            args.output_base, "classification", output_dir=args.output_dir
        )
        with FileOutputCSV(summary_outfile) as out_fp:
            tax_utils.write_summary(
                classified_results,
                out_fp,
                limit_float_decimals=limit_float,
                classification=True,
            )

    # write summarized output in human-readable format
    if "human" in args.output_format:
        summary_outfile, limit_float = make_outfile(
            args.output_base, "human", output_dir=args.output_dir
        )

        with FileOutput(summary_outfile) as out_fp:
            tax_utils.write_human_summary(
                query_gather_results,
                out_fp,
                args.rank or "species",
                classification=True,
            )

    # The following require a single rank:
    # note: interactive krona can handle mult ranks, do we want to enable?
    if "krona" in args.output_format:
        krona_results, header = tax_utils.format_for_krona(
            query_gather_results=query_gather_results,
            rank=args.rank,
            classification=True,
        )
        krona_outfile, limit_float = make_outfile(
            args.output_base, "krona", output_dir=args.output_dir
        )
        with FileOutputCSV(krona_outfile) as out_fp:
            tax_utils.write_krona(header, krona_results, out_fp)

    if "lineage_csv" in args.output_format:
        lineage_outfile, _ = make_outfile(
            args.output_base, "lineage_csv", output_dir=args.output_dir
        )
        lineage_results = []
        header = None
        for q_res in query_gather_results:
            if not header:
                ranks = list(q_res.ranks)
                if (
                    "strain" in ranks
                ):  # maintains prior functionality.. but we could keep strain now, i think?
                    ranks.remove("strain")
                header = ["ident", *ranks]
            lineageD = q_res.classification_result.as_lineage_dict(
                q_res.query_info, ranks
            )
            lineage_results.append(lineageD)
        with FileOutputCSV(lineage_outfile) as out_fp:
            tax_utils.write_output(header, lineage_results, out_fp)

    # if there was a classification error, exit with err code
    if found_error:
        if not args.force:
            sys.exit(-1)


def annotate(args):
    """
    Annotate gather results with taxonomic lineage for each match.

    Produces gather csv with lineage information as the final column.
    """

    set_quiet(args.quiet)

    try:
        # first, load taxonomic_assignments
        tax_assign = MultiLineageDB.load(
            args.taxonomy_csv,
            keep_full_identifiers=args.keep_full_identifiers,
            keep_identifier_versions=args.keep_identifier_versions,
            force=args.force,
            lins=args.lins,
            ictv=args.ictv,
        )

    except ValueError as exc:
        error(f"ERROR: {str(exc)}")
        sys.exit(-1)

    if not tax_assign:
        error(
            f"ERROR: No taxonomic assignments loaded from {','.join(args.taxonomy_csv)}. Exiting."
        )
        sys.exit(-1)

    # get csv from args
    input_csvs = tax_utils.collect_gather_csvs(
        args.gather_csv, from_file=args.from_file
    )

    # handle each gather csv separately
    for n, in_csv in enumerate(input_csvs):
        try:
            # Check for a column we can use to find lineage information:
            with FileInputCSV(in_csv) as r:
                header = r.fieldnames
                # check for empty file
                if not header:
                    raise ValueError(f"Cannot read from '{in_csv}'. Is file empty?")

                # look for the column to match with taxonomic identifier
                id_col = None
                col_options = ["name", "match_name", "ident", "accession"]
                for colname in col_options:
                    if colname in header:
                        id_col = colname
                        break

                if not id_col:
                    raise ValueError(
                        f"Cannot find taxonomic identifier column in '{in_csv}'. Tried: {', '.join(col_options)}"
                    )

                notify(
                    f"Starting annotation on '{in_csv}'. Using ID column: '{id_col}'"
                )

                # make output file for this input
                out_base = os.path.basename(in_csv.rsplit(".csv")[0])
                this_outfile, _ = make_outfile(
                    out_base, "annotate", output_dir=args.output_dir
                )

                out_header = header + ["lineage"]

                with FileOutputCSV(this_outfile) as out_fp:
                    w = csv.DictWriter(out_fp, out_header)
                    w.writeheader()

                    n = 0
                    n_missed = 0
                    for n, row in enumerate(r):
                        # find lineage and write annotated row
                        taxres = AnnotateTaxResult(
                            raw=row,
                            id_col=id_col,
                            lins=args.lins,
                            ictv=args.ictv,
                            keep_full_identifiers=args.keep_full_identifiers,
                            keep_identifier_versions=args.keep_identifier_versions,
                        )
                        taxres.get_match_lineage(
                            tax_assignments=tax_assign,
                            fail_on_missing_taxonomy=args.fail_on_missing_taxonomy,
                        )

                        if taxres.missed_ident:  # could not assign taxonomy
                            n_missed += 1
                        w.writerow(taxres.row_with_lineages())

                    rows_annotated = (n + 1) - n_missed
                    if not rows_annotated:
                        raise ValueError(
                            f"Could not annotate any rows from '{in_csv}'."
                        )
                    else:
                        notify(
                            f"Annotated {rows_annotated} of {n + 1} total rows from '{in_csv}'."
                        )

        except ValueError as exc:
            if args.force:
                notify(str(exc))
                notify("--force is set. Attempting to continue to next file.")
            else:
                error(f"ERROR: {str(exc)}")
                sys.exit(-1)


def prepare(args):
    "Combine multiple taxonomy databases into one and/or translate formats."
    notify("loading taxonomies...")
    try:
        tax_assign = MultiLineageDB.load(
            args.taxonomy_csv,
            force=args.force,
            keep_full_identifiers=args.keep_full_identifiers,
            keep_identifier_versions=args.keep_identifier_versions,
        )
    except ValueError as exc:
        error("ERROR while loading taxonomies!")
        error(str(exc))
        sys.exit(-1)

    notify(f"...loaded {len(tax_assign)} entries.")

    notify(f"saving to '{args.output}', format {args.database_format}...")
    try:
        tax_assign.save(args.output, args.database_format)
    except ValueError as exc:
        error("ERROR while saving!")
        error(str(exc))
        sys.exit(-1)

    notify("done!")


def grep(args):
    term = args.pattern
    tax_assign = MultiLineageDB.load(args.taxonomy_csv, force=args.force)

    silent = args.silent or args.count

    notify(f"searching {len(args.taxonomy_csv)} taxonomy files for '{term}'")
    if args.invert_match:
        notify(
            "-v/--invert-match specified; returning only lineages that do not match."
        )
    if args.rank:
        notify(f"limiting matches to {args.rank} level")

    # build the search pattern
    pattern = args.pattern
    if args.ignore_case:
        pattern = re.compile(pattern, re.IGNORECASE)
    else:
        pattern = re.compile(pattern)

    # determine if lineage matches.
    def find_pattern(lineage, select_rank):
        for lp in lineage:
            if select_rank is None or lp.rank == select_rank:
                if pattern.search(lp.name):
                    return True
        return False

    if args.invert_match:

        def search_pattern(l, r):
            return not find_pattern(l, r)
    else:
        search_pattern = find_pattern

    match_ident = []
    for ident, lineage in tax_assign.items():
        if search_pattern(lineage, args.rank):
            match_ident.append((ident, lineage))

    if silent:
        notify(f"found {len(match_ident)} matches.")
        notify("(no matches will be saved because of --silent/--count")
    else:
        with FileOutputCSV(args.output) as fp:
            w = csv.writer(fp)
            w.writerow(["ident"] + list(RankLineageInfo().taxlist[:-1]))
            for ident, lineage in sorted(match_ident):
                w.writerow([ident] + [x.name for x in lineage])

        notify(
            f"found {len(match_ident)} matches; saved identifiers to picklist file '{args.output}'"
        )


def summarize(args):
    "Summarize multiple taxonomy databases."
    notify("loading taxonomies...")
    try:
        tax_assign = MultiLineageDB.load(
            args.taxonomy_files,
            force=args.force,
            keep_full_identifiers=args.keep_full_identifiers,
            keep_identifier_versions=args.keep_identifier_versions,
            lins=args.lins,
            ictv=args.ictv,
        )
    except ValueError as exc:
        error("ERROR while loading taxonomies!")
        error(str(exc))
        sys.exit(-1)

    notify(f"...loaded {len(tax_assign)} entries.")

    print_results(f"number of distinct taxonomic lineages: {len(tax_assign)}")

    # count the number of distinct lineage names seen
    rank_counts = defaultdict(int)
    name_seen = set()
    for v in tax_assign.values():
        for vv in v:
            name = vv.name
            rank = vv.rank
            if name not in name_seen:
                rank_counts[rank] += 1
                name_seen.add(name)

    rank_count_items = list(rank_counts.items())
    rank_count_items.sort(key=lambda x: x[1])
    for rank, count in rank_count_items:
        rank_name_str = f"{rank}:"
        print_results(f"rank {rank_name_str:<20s} {count} distinct taxonomic lineages")

    if args.output_lineage_information:
        notify("now calculating detailed lineage counts...")
        lineage_counts = Counter()
        for v in tax_assign.values():
            tup = v
            while tup:
                lineage_counts[tup] += 1
                tup = tup[:-1]
        notify("...done!")

        with FileOutputCSV(args.output_lineage_information) as fp:
            w = csv.writer(fp)
            w.writerow(["rank", "lineage_count", "lineage"])

            # output in order of most common
            for lineage, count in lineage_counts.most_common():
                rank = lineage[-1].rank
                if args.lins:
                    inf = LINLineageInfo(lineage=lineage)
                elif args.ictv:
                    inf = ICTVRankLineageInfo(lineage=lineage)
                else:
                    inf = RankLineageInfo(lineage=lineage)
                lin = inf.display_lineage()
                w.writerow([rank, str(count), lin])

        n = len(lineage_counts)
        notify(f"saved {n} lineage counts to '{args.output_lineage_information}'")


def main(arglist=None):
    args = sourmash.cli.get_parser().parse_args(arglist)
    submod = getattr(sourmash.cli.sig, args.subcmd)
    mainmethod = getattr(submod, "main")
    return mainmethod(args)


if __name__ == "__main__":
    main(sys.argv)



"""create LCA database"""

from sourmash.cli.utils import add_ksize_arg, add_moltype_args, add_picklist_args


def subparser(subparsers):
    subparser = subparsers.add_parser("index")
    subparser.add_argument("csv", help="taxonomy spreadsheet")
    subparser.add_argument("lca_db_out", help="output database name")
    subparser.add_argument(
        "signatures",
        nargs="*",
        help="signatures or directory of signatures to index (optional if provided via --from-file)",
    )
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    subparser.add_argument("--scaled", metavar="S", default=10000, type=float)
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-d", "--debug", action="store_true", help="output debugging output"
    )
    subparser.add_argument(
        "-C",
        "--start-column",
        metavar="C",
        default=2,
        type=int,
        help="column at which taxonomic assignments start; default=2",
    )
    subparser.add_argument(
        "--tabs",
        action="store_true",
        help="input spreadsheet is tab-delimited; default is commas",
    )
    subparser.add_argument(
        "--no-headers",
        action="store_true",
        help="no headers present in taxonomy spreadsheet",
    )
    subparser.add_argument(
        "--split-identifiers",
        action="store_true",
        help="split names in signatures on whitespace",
    )
    subparser.add_argument(
        "--keep-identifier-versions",
        action="store_true",
        help="do not remove accession versions",
    )
    subparser.add_argument("-f", "--force", action="store_true")
    subparser.add_argument("--report", help="output a report on anomalies, if any")
    subparser.add_argument(
        "--require-taxonomy",
        action="store_true",
        help="ignore signatures with no taxonomy entry",
    )
    subparser.add_argument(
        "--fail-on-missing-taxonomy",
        action="store_true",
        help="fail quickly if taxonomy is not available for an identifier",
    )
    subparser.add_argument(
        "-F",
        "--database-format",
        help="format of output database; default is 'json')",
        default="json",
        choices=["json", "sql"],
    )

    add_ksize_arg(subparser, default=31)
    add_moltype_args(subparser)
    add_picklist_args(subparser)


def main(args):
    import sourmash

    return sourmash.lca.command_index.index(args)



import abc
from io import BytesIO
import os
import shutil
import sys
import tarfile
from tempfile import NamedTemporaryFile
import zipfile
from abc import ABC
from pathlib import Path
import warnings

from ._lowlevel import ffi, lib
from .utils import RustObject, rustcall, decode_str
from .minhash import to_bytes


class Storage(ABC):
    @abc.abstractmethod
    def save(self, path, content, *, overwrite=False):
        pass

    @abc.abstractmethod
    def load(self, path):
        pass

    def list_sbts(self):
        return []

    def init_args(self):
        return {}

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        pass

    def can_open(self, location):
        return False


class FSStorage(Storage):
    def __init__(self, location, subdir, make_dirs=True):
        self.location = location
        self.subdir = subdir

        if make_dirs:
            fullpath = os.path.join(location, subdir)
            if not os.path.exists(fullpath):
                os.makedirs(fullpath)

    def init_args(self):
        return {"path": self.subdir}

    def save(self, path, content, overwrite=False):
        "Save a node/leaf."
        newpath = path
        fullpath = os.path.join(self.location, self.subdir, path)

        if os.path.exists(fullpath):
            # check for content, if same return path,
            with open(fullpath, "rb") as f:
                old_content = f.read()
                if old_content == content:
                    return path

            if overwrite:
                pass  #  fine to overwrite file!
            else:
                # different content, need to find new path to save
                newpath = None
                n = 0
                while newpath is None:
                    testpath = f"{fullpath}_{n}"
                    if os.path.exists(testpath):
                        n += 1
                    else:
                        # testpath is available, use it as newpath
                        newpath = f"{path}_{n}"

        fullpath = os.path.join(self.location, self.subdir, newpath)
        with open(fullpath, "wb") as f:
            f.write(content)

        return newpath

    def load(self, path):
        path = Path(self.location) / self.subdir / path
        return path.read_bytes()


class ZipStorage(RustObject, Storage):
    __dealloc_func__ = lib.zipstorage_free

    def __init__(self, path, *, mode="r"):
        if mode == "w":
            self.__inner = _RwZipStorage(path)
        else:
            self.__inner = None
            path = os.path.abspath(path)
            self._objptr = rustcall(lib.zipstorage_new, to_bytes(path), len(path))

    @staticmethod
    def can_open(location):
        return zipfile.is_zipfile(location)

    @property
    def path(self):
        if self.__inner:
            return self.__inner.path
        return decode_str(self._methodcall(lib.zipstorage_path))

    @property
    def subdir(self):
        if self.__inner:
            return self.__inner.subdir
        return decode_str(self._methodcall(lib.zipstorage_subdir))

    @subdir.setter
    def subdir(self, value):
        if self.__inner:
            self.__inner.subdir = value
        else:
            self._methodcall(lib.zipstorage_set_subdir, to_bytes(value), len(value))

    def _filenames(self):
        if not self._objptr:
            return self.__inner._filenames()

        size = ffi.new("uintptr_t *")
        paths_ptr = self._methodcall(lib.zipstorage_filenames, size)
        size = size[0]

        paths = []
        for i in range(size):
            path = decode_str(paths_ptr[i][0])
            paths.append(path)

        return paths

    def save(self, path, content, *, overwrite=False, compress=False):
        if self.__inner:
            return self.__inner.save(
                path, content, overwrite=overwrite, compress=compress
            )
        raise NotImplementedError()

    def load(self, path):
        if not self._objptr:
            return self.__inner.load(path)

        try:
            size = ffi.new("uintptr_t *")
            rawbuf = self._methodcall(
                lib.zipstorage_load, to_bytes(path), len(path), size
            )
            size = size[0]

            rawbuf = ffi.gc(rawbuf, lambda o: lib.nodegraph_buffer_free(o, size), size)
            buf = ffi.buffer(rawbuf, size)

            # TODO: maybe avoid the [:] here, it triggers a copy...
            return buf[:]
        except ValueError:
            raise FileNotFoundError(path)

    def list_sbts(self):
        if self.__inner:
            return self.__inner.list_sbts()

        size = ffi.new("uintptr_t *")
        paths_ptr = self._methodcall(lib.zipstorage_list_sbts, size)
        size = size[0]

        paths = []
        for i in range(size):
            path = decode_str(paths_ptr[i][0])
            paths.append(path)

        return paths

    def init_args(self):
        return {"path": self.path}

    def flush(self):
        if self.__inner:
            self.__inner.flush()

    def close(self):
        if self.__inner:
            self.__inner.close()

    @staticmethod
    def can_open(location):
        return zipfile.is_zipfile(location)


class _RwZipStorage(Storage):
    def __init__(self, path):
        self.path = os.path.abspath(path)

        dirname = os.path.dirname(self.path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        self.bufferzip = None

        # Turns out we can't delete/modify an entry in a zipfile easily,
        # so we need to check some things:
        if not os.path.exists(self.path):
            # If the file doesn't exist open it in write mode.
            self.zipfile = zipfile.ZipFile(
                path, mode="w", compression=zipfile.ZIP_STORED
            )
        else:
            # If it exists, open it in read mode and prepare a buffer for
            # new/duplicated items. During close() there are checks to see
            # how the original file needs to be updated (append new items,
            # deal with duplicates, and so on)
            self.zipfile = zipfile.ZipFile(path, "r")
            self.bufferzip = zipfile.ZipFile(BytesIO(), mode="w")

        self.subdir = ""
        subdirs = [f for f in self.zipfile.namelist() if f.endswith("/")]
        if len(subdirs) == 1:
            self.subdir = subdirs[0]

    def _filenames(self):
        return [info.filename for info in self.zipfile.infolist()]

    def _content_matches(self, zf, path, content):
        info = zf.getinfo(path)
        entry_content = zf.read(info)
        if entry_content == content:
            return True
        return False

    def _generate_filename(self, zf, path, content):
        try:
            matches = self._content_matches(zf, path, content)
            if matches:
                return path, False
        except KeyError:
            # entry not there yet, use that path
            return path, True

        # content does not match - generate new path based on path
        newpath = None
        n = 0
        while newpath is None:
            testpath = f"{path}_{n}"
            try:
                matches = self._content_matches(zf, testpath, content)
                if matches:
                    return testpath, False
                else:
                    n += 1
            except KeyError:
                return testpath, True

        assert 0  # should never get here!

    def _write_to_zf(self, zf, path, content, *, compress=False):
        compress_type = zipfile.ZIP_STORED
        if compress:
            compress_type = zipfile.ZIP_DEFLATED

        # save to zipfile
        zf.writestr(path, content, compress_type=compress_type)

        # set permissions
        zi = zf.getinfo(path)
        perms = 0o444 << 16  # give a+r access
        if path.endswith("/"):
            perms = 0o755 << 16  # directories get u+rwx, a+rx
        zi.external_attr = perms

    def save(self, path, content, *, overwrite=False, compress=False):
        # ignore UserWarnings for duplicate filenames.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # First try to save to self.zipfile, if it is not writable
            # or would introduce duplicates then try to save it in the buffer
            if overwrite:
                newpath = path
                do_write = True
            else:
                newpath, do_write = self._generate_filename(self.zipfile, path, content)
            if do_write:
                try:
                    self._write_to_zf(self.zipfile, newpath, content, compress=compress)
                except (ValueError, RuntimeError):
                    # Can't write in the zipfile, write in buffer instead
                    # CTB: do we need to generate a new filename wrt to the
                    # bufferzip, too? Not sure this code is working as intended...
                    if self.bufferzip:
                        self._write_to_zf(
                            self.bufferzip, newpath, content, compress=compress
                        )
                    else:
                        # Throw error, can't write the data
                        raise ValueError("can't write data")

        return newpath

    def _load_from_zf(self, zf, path):
        # we repeat these steps for self.zipfile and self.bufferzip,
        # so better to have an auxiliary method
        try:
            return zf.read(path)
        except KeyError:
            path = os.path.join(self.subdir, path)
            return zf.read(path)

    def load(self, path):
        try:
            return self._load_from_zf(self.zipfile, path)
        except KeyError:
            if self.bufferzip:
                return self._load_from_zf(self.bufferzip, path)
            else:
                raise FileNotFoundError(path)

    def close(self):
        # TODO: this is not ideal; checking for zipfile.fp is looking at
        # internal implementation details from CPython...

        # might not have self.zipfile if was invalid zipfile and __init__
        # failed.
        if hasattr(self, "zipfile"):
            if self.zipfile is not None or self.bufferzip is not None:
                self.flush(keep_closed=True)
                self.zipfile.close()
                self.zipfile = None

    def flush(self, *, keep_closed=False):
        # This is a bit complicated, but we have to deal with new data
        # (if the original zipfile is read-only) and possible duplicates.

        if self.bufferzip is None:
            # The easy case: close (to force flushing) and reopen the zipfile
            if self.zipfile is not None:
                self.zipfile.close()
                if not keep_closed:
                    self.zipfile = zipfile.ZipFile(
                        self.path, mode="a", compression=zipfile.ZIP_STORED
                    )
        else:
            # The complicated one. Need to consider:
            # - Is there data in the buffer?
            # - If there is, is any of it
            #    * duplicated?
            #    * new data?
            buffer_names = set(self.bufferzip.namelist())
            zf_names = set(self.zipfile.namelist())
            if buffer_names:
                new_data = buffer_names - zf_names
                duplicated = buffer_names & zf_names

                if duplicated:
                    # bad news, need to create new file...
                    # create a temporary file to write the final version,
                    # which will be copied to the right place later.
                    tempfile = NamedTemporaryFile(delete=False)
                    final_file = zipfile.ZipFile(tempfile, mode="w")
                    all_data = buffer_names.union(zf_names)

                    for item in all_data:
                        if item in duplicated or item in buffer_names:
                            # we prioritize writing data from the buffer to the
                            # final file
                            self._write_to_zf(
                                final_file, item, self.bufferzip.read(item)
                            )
                        else:
                            # it is only in the zipfile, so write from it
                            self._write_to_zf(final_file, item, self.zipfile.read(item))

                    # close the files, remove the old one and copy the final
                    # file to the right place.
                    self.zipfile.close()
                    final_file.close()
                    os.unlink(self.path)
                    shutil.move(tempfile.name, self.path)
                    if not keep_closed:
                        self.zipfile = zipfile.ZipFile(
                            self.path, mode="a", compression=zipfile.ZIP_STORED
                        )
                elif new_data:
                    # Since there is no duplicated data, we can
                    # reopen self.zipfile in append mode and write the new data
                    self.zipfile.close()
                    if keep_closed:
                        raise Exception("unexpected error")
                    else:
                        zf = zipfile.ZipFile(
                            self.path, mode="a", compression=zipfile.ZIP_STORED
                        )
                    for item in new_data:
                        self._write_to_zf(zf, item, self.bufferzip.read(item))
                    self.zipfile = zf
            # finally, close the buffer and release memory
            self.bufferzip.close()
            self.bufferzip = None

    def list_sbts(self):
        return [f for f in self.zipfile.namelist() if f.endswith(".sbt.json")]

    def __del__(self):
        self.close()


class IPFSStorage(Storage):
    def __init__(self, pin_on_add=True, **kwargs):
        import ipfshttpclient

        self.ipfs_args = kwargs
        self.pin_on_add = pin_on_add
        self.api = ipfshttpclient.connect(**self.ipfs_args)

    def save(self, path, content, *, overwrite=False):
        new_obj = self.api.add_bytes(content)
        if self.pin_on_add:
            self.api.pin.add(new_obj)
        return new_obj

        # TODO: the above solution is quick and dirty.
        # we actually want something more organized,
        # like putting all the generated objects inside the same dir.
        # Check this call using the files API for an example.
        # api.files_write("/test/file", io.BytesIO(b"hi"), create=True)
        #
        # This is also required to bring the IPFSStorage closer to what the
        # ZipStorage is doing now.

    def load(self, path):
        return self.api.cat(path)

    def init_args(self):
        return self.ipfs_args

    def __exit__(self, type, value, traceback):
        # TODO: do nothing for now,
        # but we actually want something more organized,
        # like putting all the generated objects inside the same dir.
        # Use the files API,
        # add files without flush(),
        # and then flush it here?
        pass


class RedisStorage(Storage):
    def __init__(self, **kwargs):
        import redis

        self.redis_args = kwargs
        self.conn = redis.Redis(**self.redis_args)

    def save(self, path, content, *, overwrite=False):
        if not isinstance(content, bytes):
            content = bytes(content)
        self.conn.set(path, content)
        return path

    def load(self, path):
        return self.conn.get(path)

    def init_args(self):
        # TODO: do we want to remove stuff like password from here?
        return self.redis_args

    def __exit__(self, type, value, traceback):
        pass



#! /usr/bin/env python
"""
Use the MurmurHash library mmh3 and separate Python code to calculate
a MinHash signature for translated protein sequence, as a way to do an
external check on our C++ implementation.

The output of this is used in test_sourmash.py to verify our C++ code.
"""

dna_to_aa = {
    "TTT": "F",
    "TTC": "F",
    "TTA": "L",
    "TTG": "L",
    "TCT": "S",
    "TCC": "S",
    "TCA": "S",
    "TCG": "S",
    "TAT": "Y",
    "TAC": "Y",
    "TAA": "*",
    "TAG": "*",
    "TGA": "*",
    "TGT": "C",
    "TGC": "C",
    "TGG": "W",
    "CTT": "L",
    "CTC": "L",
    "CTA": "L",
    "CTG": "L",
    "CCT": "P",
    "CCC": "P",
    "CCA": "P",
    "CCG": "P",
    "CAT": "H",
    "CAC": "H",
    "CAA": "Q",
    "CAG": "Q",
    "CGT": "R",
    "CGC": "R",
    "CGA": "R",
    "CGG": "R",
    "ATT": "I",
    "ATC": "I",
    "ATA": "I",
    "ATG": "M",
    "ACT": "T",
    "ACC": "T",
    "ACA": "T",
    "ACG": "T",
    "AAT": "N",
    "AAC": "N",
    "AAA": "K",
    "AAG": "K",
    "AGT": "S",
    "AGC": "S",
    "AGA": "R",
    "AGG": "R",
    "GTT": "V",
    "GTC": "V",
    "GTA": "V",
    "GTG": "V",
    "GCT": "A",
    "GCC": "A",
    "GCA": "A",
    "GCG": "A",
    "GAT": "D",
    "GAC": "D",
    "GAA": "E",
    "GAG": "E",
    "GGT": "G",
    "GGC": "G",
    "GGA": "G",
    "GGG": "G",
}


__complementTranslation = {"A": "T", "C": "G", "G": "C", "T": "A", "N": "N"}


def complement(s):
    """
    Return complement of 's'.
    """
    c = "".join(__complementTranslation[n] for n in s)
    return c


def reverse(s):
    """
    Return reverse of 's'.
    """
    r = "".join(reversed(s))
    return r


def peptides(seq, start):
    for i in range(start, len(seq), 3):
        yield dna_to_aa.get(seq[i : i + 3], "X")


def translate(seq):
    for i in range(3):
        pep = peptides(seq, i)
        yield "".join(pep)

    revcomp = reverse(complement(seq))
    for i in range(3):
        pep = peptides(revcomp, i)
        yield "".join(pep)


def kmers(seq, k):
    for start in range(len(seq) - k + 1):
        yield seq[start : start + k]


###

K = 21

import sys
import screed
import mmh3
import sourmash

print("imported sourmash:", sourmash, file=sys.stderr)
import sourmash.signature

record = next(iter(screed.open(sys.argv[1])))
print("loaded", record.name, file=sys.stderr)

mh = sourmash.MinHash(ksize=K, n=500, is_protein=True)
prot_ksize = int(K / 3)

for trans in translate(record.sequence):
    for kmer in kmers(trans, prot_ksize):
        hash = mmh3.hash64(kmer, seed=42)[0]

        # convert to unsigned int if negative
        if hash < 0:
            hash += 2**64

        mh.add_hash(hash)

s = sourmash.signature.SourmashSignature("", mh, name=record.name)
print(sourmash.signature.save_signatures([s]))



from tempfile import NamedTemporaryFile

import pytest

from sourmash.nodegraph import (
    Nodegraph,
    extract_nodegraph_info,
    calc_expected_collisions,
)

import sourmash_tst_utils as utils


def test_nodegraph_to_khmer_basic():
    pytest.importorskip("khmer")

    ng_file = utils.get_test_data(".sbt.v3/internal.0")

    sourmash_ng = Nodegraph.load(ng_file)
    khmer_sm_ng = sourmash_ng.to_khmer_nodegraph()

    assert sourmash_ng.ksize() == khmer_sm_ng.ksize()


def test_nodegraph_khmer_compare():
    khmer = pytest.importorskip("khmer")

    khmer_ng = khmer.Nodegraph(3, 23, 6)
    khmer_ng.count("ACG")
    khmer_ng.count("TTA")
    khmer_ng.count("CGA")

    sm_ng = Nodegraph(3, 23, 6)
    sm_ng.count("ACG")
    sm_ng.count("TTA")
    sm_ng.count("CGA")

    assert sm_ng.ksize() == khmer_ng.ksize()
    assert sm_ng.hashsizes() == khmer_ng.hashsizes()
    assert sm_ng.get("ACG")
    assert sm_ng.get("TTA")
    assert sm_ng.get("CGA")

    assert khmer_ng.get("ACG")
    assert khmer_ng.get("TTA")
    assert khmer_ng.get("CGA")


def test_nodegraph_same_file():
    khmer = pytest.importorskip("khmer")
    try:
        load_nodegraph = khmer.load_nodegraph
    except AttributeError:
        load_nodegraph = khmer.Nodegraph.load

    ng_file = utils.get_test_data(".sbt.v3/internal.0")
    with open(ng_file, "rb") as f:
        ng_data = f.read()

    sourmash_ng = Nodegraph.load(ng_file)
    khmer_sm_ng = sourmash_ng.to_khmer_nodegraph()

    khmer_ng = load_nodegraph(ng_file)

    with (
        NamedTemporaryFile() as f1,
        NamedTemporaryFile() as f2,
        NamedTemporaryFile() as f3,
    ):
        sourmash_ng.save(f1.name)
        khmer_sm_ng.save(f2.name)
        khmer_ng.save(f3.name)

        f1.seek(0)
        sm_data = f1.read()

        f2.seek(0)
        kh_sm_data = f2.read()

        f3.seek(0)
        kh_data = f3.read()

        assert extract_nodegraph_info(f1.name) == extract_nodegraph_info(f2.name)
        assert extract_nodegraph_info(f3.name) == extract_nodegraph_info(f2.name)
        assert extract_nodegraph_info(f1.name) == extract_nodegraph_info(f3.name)

        assert sm_data == kh_data
        assert sm_data == kh_sm_data

        assert ng_data == sm_data
        assert ng_data == kh_data
        assert ng_data == kh_sm_data


def test_nodegraph_expected_collisions():
    ng_file = utils.get_test_data(".sbt.v3/internal.0")

    sourmash_ng = Nodegraph.load(ng_file)

    assert calc_expected_collisions(sourmash_ng) == 3.412442571740036e-07


def test_nodegraph_expected_collisions_error():
    ng_file = utils.get_test_data(".sbt.v3/internal.0")

    sourmash_ng = Nodegraph.load(ng_file)

    with pytest.raises(SystemExit):
        calc_expected_collisions(sourmash_ng, max_false_pos=1e-8)



"""collect manifest information across many files"""

usage = """

    sourmash sig collect <filenames> -o all.sqlmf

This will collect manifests from across many files and save the information
into a standalone manifest database.

By default, 'sig collect' requires a pre-existing manifest for collections;
this prevents potentially slow manifest rebuilding. You
can turn this check off with '--no-require-manifest'.

"""

from sourmash.cli.utils import (
    add_moltype_args,
    add_ksize_arg,
    add_picklist_args,
    add_pattern_args,
)


def subparser(subparsers):
    subparser = subparsers.add_parser("collect", usage=usage)
    subparser.add_argument("locations", nargs="*", help="locations of input signatures")
    subparser.add_argument("-o", "--output", help="manifest output file", required=True)
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-d", "--debug", action="store_true", help="provide debugging output"
    )
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    subparser.add_argument(
        "--no-require-manifest",
        help="do not require a manifest; generate dynamically if needed",
        action="store_true",
    )
    subparser.add_argument(
        "-F",
        "--manifest-format",
        help="format of manifest output file; default is 'csv')",
        default="sql",
        choices=["csv", "sql"],
    )

    subparser.add_argument(
        "--merge-previous",
        action="store_true",
        help="merge new manifests into existing",
    )
    subparser.add_argument(
        "--abspath",
        "--use-absolute-paths",
        help="convert all locations to absolute paths",
        action="store_true",
    )
    subparser.add_argument(
        "--no-abspath",
        help="do not convert all locations to absolute paths",
        action="store_false",
        dest="abspath",
    )
    subparser.add_argument(
        "--relpath",
        "--use-relative-paths",
        help="convert all locations to paths relative to the output manifest",
        action="store_true",
    )
    subparser.add_argument(
        "--no-relpath",
        help="do not convert all locations to paths relative to the output manifest",
        action="store_false",
        dest="relpath",
    )

    add_ksize_arg(subparser)
    add_moltype_args(subparser)


def main(args):
    import sourmash

    return sourmash.sig.__main__.collect(args)



"""summarize mixture"""


def subparser(subparsers):
    subparser = subparsers.add_parser("summarize")
    subparser.add_argument(
        "--db", nargs="+", action="append", help="one or more LCA databases to use"
    )
    subparser.add_argument(
        "--query",
        nargs="*",
        default=[],
        action="append",
        help="one or more signature files to use as queries",
    )
    subparser.add_argument(
        "--query-from-file", help="file containing list of signature files to query"
    )
    subparser.add_argument(
        "--threshold",
        metavar="T",
        type=int,
        default=5,
        help="minimum number of hashes to require for a match",
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        help="file to which CSV output will be written",
    )
    subparser.add_argument(
        "--scaled", metavar="FLOAT", type=float, help="scaled value to downsample to"
    )

    subparser.add_argument(
        "--ignore-abundance",
        action="store_true",
        help="ignore hash abundances in query signatures do not weight results",
    )

    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-d", "--debug", action="store_true", help="output debugging output"
    )


def main(args):
    import sourmash

    return sourmash.lca.command_summarize.summarize_main(args)



"""combine multiple taxonomy databases into one."""

usage = """

    sourmash tax prepare --taxonomy-csv <taxonomy_file> [ ... ] -o <output>

The 'tax prepare' command reads in one or more taxonomy databases
and saves them into a new database. It can be used to combine databases
in the desired order, as well as output different database formats.

Please see the 'tax prepare' documentation for more details:
  https://sourmash.readthedocs.io/en/latest/command-line.html#sourmash-tax-prepare-prepare-and-or-combine-taxonomy-files
"""

import sourmash
from sourmash.logging import notify, print_results, error


def subparser(subparsers):
    subparser = subparsers.add_parser("prepare", usage=usage)
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-t",
        "--taxonomy-csv",
        "--taxonomy",
        metavar="FILE",
        nargs="+",
        required=True,
        action="extend",
        help="database lineages",
    )
    subparser.add_argument(
        "-o",
        "--output",
        required=True,
        help="output file",
    )
    subparser.add_argument(
        "-F",
        "--database-format",
        help="format of output file; default is 'sql')",
        default="sql",
        choices=["csv", "sql"],
    )
    subparser.add_argument(
        "--keep-full-identifiers",
        action="store_true",
        help="do not split identifiers on whitespace",
    )
    subparser.add_argument(
        "--keep-identifier-versions",
        action="store_true",
        help="after splitting identifiers, do not remove accession versions",
    )
    subparser.add_argument(
        "--fail-on-missing-taxonomy",
        action="store_true",
        help="fail quickly if taxonomy is not available for an identifier",
    )
    subparser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="continue past errors in file and taxonomy loading",
    )


def main(args):
    return sourmash.tax.__main__.prepare(args)



"""extract one or more signatures"""

usage = """

### `sourmash signature extract` - extract signatures from a collection

Extract the specified signature(s) from a collection of signatures.

For example,

sourmash signature extract *.sig -k 21 --dna -o extracted.sig

will extract all nucleotide signatures calculated at k=21 from all
.sig files in the current directory.

There are currently two other useful selectors for `extract`: you can specify
(part of) an md5sum, as output in the CSVs produced by `search` and `gather`;
and you can specify (part of) a name.

For example,

sourmash signature extract tests/test-data/*.fa.sig --md5 09a0869

will extract the signature from `47.fa.sig` which has an md5sum of
`09a08691ce52952152f0e866a59f6261`; and 

sourmash signature extract tests/test-data/*.fa.sig --name NC_009665

will extract the same signature, which has an accession number of
`NC_009665.1`.

#### Using picklists with `sourmash sig extract`

As of sourmash 4.2.0, `extract` also supports picklists, a feature by
which you can select signatures based on values in a CSV file. See
[the command line docs](https://sourmash.readthedocs.io/en/latest/command-line.html) for more information.

"""

from sourmash.cli.utils import (
    add_moltype_args,
    add_ksize_arg,
    add_picklist_args,
    add_pattern_args,
)


def subparser(subparsers):
    subparser = subparsers.add_parser("extract", description=__doc__, usage=usage)
    subparser.add_argument("signatures", nargs="*")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        help="output signature to this file (default stdout)",
        default="-",
    )
    subparser.add_argument(
        "--md5",
        default=None,
        help="select signatures whose md5 contains this substring",
    )
    subparser.add_argument(
        "--name",
        default=None,
        help="select signatures whose name contains this substring",
    )
    subparser.add_argument(
        "-f", "--force", action="store_true", help="try to load all files as signatures"
    )
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_pattern_args(subparser)
    add_picklist_args(subparser)


def main(args):
    import sourmash

    return sourmash.sig.__main__.extract(args)



"""annotate gather results with taxonomy information"""

usage = """

    sourmash tax annotate --gather-csv <gather_csv> [ ... ] --taxonomy-csv <taxonomy_csv> [ ... ]

The 'tax annotate' command reads in gather results CSVs and annotates them
 with taxonomic information.

By default, 'tax annotate' produces a gather CSV with an additional 'lineage'
 column containing the taxonomic information for each database match.

Please see the 'tax annotate' documentation for more details:
  https://sourmash.readthedocs.io/en/latest/command-line.html#sourmash-tax-annotate-annotates-gather-output-with-taxonomy
"""

import sourmash
from sourmash.logging import notify, print_results, error


def subparser(subparsers):
    subparser = subparsers.add_parser("annotate", aliases=["annotate"], usage=usage)
    subparser.add_argument(
        "-g",
        "--gather-csv",
        nargs="*",
        default=[],
        action="extend",
        help="CSV output files from sourmash gather",
    )
    subparser.add_argument(
        "--from-file",
        metavar="FILE",
        default=None,
        help="input many gather results as a text file, with one gather CSV per line",
    )
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-t",
        "--taxonomy-csv",
        "--taxonomy",
        metavar="FILE",
        nargs="*",
        required=True,
        action="extend",
        help="database lineages CSV",
    )
    subparser.add_argument(
        "-o", "--output-dir", default="", help="directory for output files"
    )
    subparser.add_argument(
        "--keep-full-identifiers",
        action="store_true",
        help="do not split identifiers on whitespace",
    )
    subparser.add_argument(
        "--keep-identifier-versions",
        action="store_true",
        help="after splitting identifiers, do not remove accession versions",
    )
    subparser.add_argument(
        "--fail-on-missing-taxonomy",
        action="store_true",
        help="fail quickly if taxonomy is not available for an identifier",
    )
    subparser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="continue past errors in file and taxonomy loading",
    )
    subparser.add_argument(
        "--lins",
        "--lin-taxonomy",
        action="store_true",
        default=False,
        help="use LIN taxonomy in place of standard taxonomic ranks.  Note that the taxonomy CSV must contain LIN lineage information.",
    )
    subparser.add_argument(
        "--ictv",
        "--ictv-taxonomy",
        action="store_true",
        default=False,
        help="use ICTV taxonomy in place of standard taxonomic ranks.  Note that the taxonomy CSV must contain ICTV ranks.",
    )


def main(args):
    if not args.gather_csv and not args.from_file:
        raise ValueError(
            "No gather CSVs found! Please input via '-g' or '--from-file'."
        )
    return sourmash.tax.__main__.annotate(args)



"""
Tests for the picklist API.
"""

import pytest
import sourmash

import copy

import sourmash_tst_utils as utils
from sourmash import picklist
from sourmash.picklist import SignaturePicklist
from sourmash.index import LinearIndex, MultiIndex
from sourmash.index.sqlite_index import SqliteIndex


def test_load_empty_picklist_fail():
    empty = utils.get_test_data("picklist/empty.csv")

    pl = SignaturePicklist("manifest", pickfile=empty)
    with pytest.raises(ValueError):
        pl.load(allow_empty=False)


def test_load_empty_picklist_allow():
    empty = utils.get_test_data("picklist/empty.csv")

    pl = SignaturePicklist("manifest", pickfile=empty)
    pl.load(allow_empty=True)


def test_dup_md5_picked(runtmp):
    # load a sig, duplicate, and see if a picklist gets the right one
    sig47 = utils.get_test_data("47.fa.sig")
    ss = sourmash.load_file_as_signatures(sig47)
    sig = list(ss)[0]

    # save a manifest with one entry
    xl = LinearIndex([sig])
    ml = MultiIndex.load([xl], [None], None)

    print(ml.manifest.rows)
    assert len(ml.manifest) == 1

    mf_csv = runtmp.output("select.csv")
    ml.manifest.write_to_filename(mf_csv)

    # now make an index to select against, with an identical signature
    # (but diff name)
    new_sig = sig.to_mutable()
    new_sig.name = "foo"
    xl = LinearIndex([sig, new_sig])
    ml2 = MultiIndex.load([xl], [None], None)

    assert len(ml2) == 2

    # create a picklist...
    pl = SignaturePicklist("manifest", pickfile=mf_csv)
    print(pl.load())
    print("loaded:", len(pl.pickset))

    # use in select
    ml3 = ml2.select(picklist=pl)
    print("picked:", len(ml3))

    assert len(pl.pickset) == len(ml3)


def test_dup_md5_picked_mf_to_picklist(runtmp):
    # load a sig, duplicate, and see if a picklist gets the right one
    # uses an in memory picklist
    sig47 = utils.get_test_data("47.fa.sig")
    ss = sourmash.load_file_as_signatures(sig47)
    sig = list(ss)[0]

    # save a manifest with one entry
    xl = LinearIndex([sig])
    ml = MultiIndex.load([xl], [None], None)

    print(ml.manifest.rows)
    assert len(ml.manifest) == 1

    pl = ml.manifest.to_picklist()

    # now make an index to select against, with an identical signature
    # (but diff name)
    new_sig = sig.to_mutable()
    new_sig.name = "foo"
    xl = LinearIndex([sig, new_sig])
    ml2 = MultiIndex.load([xl], [None], None)

    assert len(ml2) == 2

    # use picklist in select
    ml3 = ml2.select(picklist=pl)
    print("picked:", len(ml3))

    assert len(pl.pickset) == len(ml3)


def test_dup_md5_picked_mf_to_picklist_sqlite(runtmp):
    # load a sig, duplicate, and see if a picklist gets the right one
    # use a sqlite db with its own to_picklist behavior.
    sig47 = utils.get_test_data("47.fa.sig")
    ss = sourmash.load_file_as_signatures(sig47)
    sig = list(ss)[0]

    # save a manifest with one entry
    xl = SqliteIndex.create(":memory:")
    xl.insert(sig)

    print(xl.manifest.rows)
    assert len(xl.manifest) == 1

    pl = xl.manifest.to_picklist()

    # now make an index to select against, with an identical signature
    # (but diff name)
    new_sig = sig.to_mutable()
    new_sig.name = "foo"
    xl = LinearIndex([sig, new_sig])
    ml2 = MultiIndex.load([xl], [None], None)

    assert len(ml2) == 2

    # use picklist in select
    ml3 = ml2.select(picklist=pl)
    print("picked:", len(ml3))

    assert len(pl.pickset) == len(ml3)



from glob import glob
import os
import argparse
from sourmash.logging import notify
from sourmash.sourmash_args import check_scaled_bounds, check_num_bounds


def add_moltype_args(parser):
    parser.add_argument(
        "--protein",
        dest="protein",
        action="store_true",
        help="choose a protein signature; by default, a nucleotide signature is used",
    )
    parser.add_argument(
        "--no-protein",
        dest="protein",
        action="store_false",
        help="do not choose a protein signature",
    )
    parser.set_defaults(protein=False)

    parser.add_argument(
        "--dayhoff",
        dest="dayhoff",
        action="store_true",
        help="choose Dayhoff-encoded amino acid signatures",
    )
    parser.add_argument(
        "--no-dayhoff",
        dest="dayhoff",
        action="store_false",
        help="do not choose Dayhoff-encoded amino acid signatures",
    )
    parser.set_defaults(dayhoff=False)

    parser.add_argument(
        "--hp",
        "--hydrophobic-polar",
        dest="hp",
        action="store_true",
        help="choose hydrophobic-polar-encoded amino acid signatures",
    )
    parser.add_argument(
        "--no-hp",
        "--no-hydrophobic-polar",
        dest="hp",
        action="store_false",
        help="do not choose hydrophobic-polar-encoded amino acid signatures",
    )
    parser.set_defaults(hp=False)

    parser.add_argument(
        "--skipm1n3",
        "--skipmer-m1n3",
        dest="skipm1n3",
        action="store_true",
        help="choose skipmer (m1n3) signatures",
    )

    parser.add_argument(
        "--no-skipm1n3",
        "--no-skipmer-m1n3",
        dest="skipm1n3",
        action="store_false",
        help="do not choose skipmer (m1n3) signatures",
    )
    parser.set_defaults(skipm1n3=False)

    parser.add_argument(
        "--skipm2n3",
        "--skipmer-m2n3",
        dest="skipm2n3",
        action="store_true",
        help="choose skipmer (m2n3) signatures",
    )
    parser.add_argument(
        "--no-skipm2n3",
        "--no-skipmer-m2n3",
        dest="skipm2n3",
        action="store_false",
        help="do not choose skipmer (m2n3) signatures",
    )
    parser.set_defaults(skipm2n3=False)

    parser.add_argument(
        "--dna",
        "--rna",
        "--nucleotide",
        dest="dna",
        default=None,
        action="store_true",
        help="choose a nucleotide signature (default: True)",
    )
    parser.add_argument(
        "--no-dna",
        "--no-rna",
        "--no-nucleotide",
        dest="dna",
        action="store_false",
        help="do not choose a nucleotide signature",
    )
    parser.set_defaults(dna=None)


def add_construct_moltype_args(parser):
    add_moltype_args(parser)
    parser.set_defaults(dna=True)


def add_ksize_arg(parser, *, default=None):
    "Add -k/--ksize to argparse parsers, with specified default."
    if default:
        message = f"k-mer size to select; default={default}"
    else:
        message = "k-mer size to select; no default."

    parser.add_argument(
        "-k",
        "--ksize",
        metavar="K",
        default=default,
        type=int,
        help=message,
    )


# https://stackoverflow.com/questions/55324449/how-to-specify-a-minimum-or-maximum-float-value-with-argparse#55410582
def range_limited_float_type(arg):
    """Type function for argparse - a float within some predefined bounds"""
    min_val = 0
    max_val = 1
    try:
        f = float(arg)
    except ValueError:
        raise argparse.ArgumentTypeError("\n\tERROR: Must be a floating point number.")
    if f < min_val or f > max_val:
        raise argparse.ArgumentTypeError(
            f"\n\tERROR: Argument must be >{str(min_val)} and <{str(max_val)}."
        )
    return f


def add_tax_threshold_arg(parser, containment_default=0.1, ani_default=None):
    parser.add_argument(
        "--containment-threshold",
        default=containment_default,
        type=range_limited_float_type,
        help=f"minimum containment threshold for classification; default={containment_default}",
    )
    parser.add_argument(
        "--ani-threshold",
        "--aai-threshold",
        default=ani_default,
        type=range_limited_float_type,
        help=f"minimum ANI threshold (nucleotide gather) or AAI threshold (protein gather) for classification; default={ani_default}",
    )


def add_picklist_args(parser):
    parser.add_argument(
        "--picklist",
        default=None,
        help="select signatures based on a picklist, i.e. 'file.csv:colname:coltype'",
    )
    parser.add_argument(
        "--picklist-require-all",
        default=False,
        action="store_true",
        help="require that all picklist values be found or else fail",
    )


def add_pattern_args(parser):
    parser.add_argument(
        "--include-db-pattern",
        default=None,
        help="search only signatures that match this pattern in name, filename, or md5",
    )
    parser.add_argument(
        "--exclude-db-pattern",
        default=None,
        help="search only signatures that do not match this pattern in name, filename, or md5",
    )


def opfilter(path):
    return not path.startswith("__") and path not in ["utils"]


def command_list(dirpath):
    paths = glob(os.path.join(dirpath, "*.py"))
    filenames = [os.path.basename(path) for path in paths]
    basenames = [
        os.path.splitext(path)[0] for path in filenames if not path.startswith("__")
    ]
    basenames = filter(opfilter, basenames)
    return sorted(basenames)


def add_scaled_arg(parser, default=None):
    parser.add_argument(
        "--scaled",
        metavar="FLOAT",
        type=check_scaled_bounds,
        help="downsample to this scaled; value should be between 100 and 1e6",
    )


def add_num_arg(parser, default=0):
    parser.add_argument(
        "-n",
        "--num-hashes",
        "--num",
        metavar="N",
        type=check_num_bounds,
        default=default,
        help="num value should be between 50 and 50000",
    )


def check_rank(args):
    """Check '--rank'/'--position'/'--lin-position' argument matches selected taxonomy."""
    standard_ranks = [
        "strain",
        "species",
        "genus",
        "family",
        "order",
        "class",
        "phylum",
        "superkingdom",
    ]
    if args.lins:
        if args.rank.isdigit():
            return str(args.rank)
        raise argparse.ArgumentTypeError(
            f"Invalid '--rank'/'--position' input: '{args.rank}'. '--lins' is specified. Rank must be an integer corresponding to a LIN position."
        )
    elif args.rank in standard_ranks:
        return args.rank
    else:
        raise argparse.ArgumentTypeError(
            f"Invalid '--rank'/'--position' input: '{args.rank}'. Please choose: 'strain', 'species', 'genus', 'family', 'order', 'class', 'phylum', 'superkingdom'"
        )


def add_rank_arg(parser):
    parser.add_argument(
        "-r",
        "--rank",
        "--position",
        "--lin-position",
        help="For non-default output formats. Classify to this rank (tax genome) or summarize taxonomy at this rank and above (tax metagenome). \
              Note that the taxonomy CSV must contain lineage information at this rank, and that LIN positions start at 0. \
              Choices: 'strain', 'species', 'genus', 'family', 'order', 'class', 'phylum', 'superkingdom' or an integer LIN position",
    )


def check_tax_outputs(
    args,
    rank_required=["krona"],
    incompatible_with_lins=None,
    use_lingroup_format=False,
):
    "Handle ouput format combinations"
    # check that rank is passed for formats requiring rank.
    if not args.rank:
        if any(x in rank_required for x in args.output_format):
            raise ValueError(
                f"Rank (--rank) is required for {', '.join(rank_required)} output formats."
            )

    if args.lins:
        # check for outputs incompatible with lins
        if incompatible_with_lins:
            if any(x in args.output_format for x in incompatible_with_lins):
                raise ValueError(
                    f"The following outputs are incompatible with '--lins': : {', '.join(incompatible_with_lins)}"
                )
        # check that lingroup file exists if needed
        if args.lingroup:
            if use_lingroup_format and "lingroup" not in args.output_format:
                args.output_format.append("lingroup")
        elif "lingroup" in args.output_format:
            raise ValueError(
                "Must provide lingroup csv via '--lingroup' in order to output a lingroup report."
            )
    elif args.lingroup or "lingroup" in args.output_format:
        raise ValueError(
            "Must enable LIN taxonomy via '--lins' in order to use lingroups."
        )

    # check that only one output format is specified if writing to stdout
    if len(args.output_format) > 1:
        if args.output_base == "-":
            raise ValueError(
                f"Writing to stdout is incompatible with multiple output formats {args.output_format}"
            )
    elif not args.output_format:
        # change to "human" for 5.0
        args.output_format = ["csv_summary"]

    return args.output_format



"""
Support for plugins to sourmash via importlib.metadata entrypoints.

Plugin entry point names:
* 'sourmash.load_from' - Index class loading.
* 'sourmash.save_to' - Signature saving.
* 'sourmash.cli_script' - command-line extension.

CTB TODO:

* consider using something other than 'name' for loader fn name. Maybe __doc__?
* try implement picklist plugin?
"""

DEFAULT_LOAD_FROM_PRIORITY = 99
DEFAULT_SAVE_TO_PRIORITY = 99

import itertools
import argparse

from .logging import debug_literal, error, notify, set_quiet

# cover for older versions of Python that don't support selection on load
# (the 'group=' below).
from importlib.metadata import entry_points

# load 'load_from' entry points. NOTE: this executes on import of this module.
try:
    _plugin_load_from = entry_points(group="sourmash.load_from")
except TypeError:
    from importlib_metadata import entry_points

    _plugin_load_from = entry_points(group="sourmash.load_from")

# load 'save_to' entry points as well.
_plugin_save_to = entry_points(group="sourmash.save_to")

# aaaaand CLI entry points:
_plugin_cli = entry_points(group="sourmash.cli_script")
_plugin_cli_once = False

###


def get_load_from_functions():
    "Load the 'load_from' plugins and yield tuples (priority, name, fn)."
    debug_literal(f"load_from plugins: {_plugin_load_from}")

    # Load each plugin,
    for plugin in _plugin_load_from:
        try:
            loader_fn = plugin.load()
        except (ModuleNotFoundError, AttributeError) as e:
            debug_literal(
                f"plugins.load_from_functions: got error loading {plugin.name}: {str(e)}"
            )
            continue

        # get 'priority' if it is available
        priority = getattr(loader_fn, "priority", DEFAULT_LOAD_FROM_PRIORITY)

        # retrieve name (which is specified by plugin?)
        name = plugin.name
        debug_literal(f"plugins.load_from_functions: got '{name}', priority={priority}")
        yield priority, name, loader_fn


def get_save_to_functions():
    "Load the 'save_to' plugins and yield tuples (priority, fn)."
    debug_literal(f"save_to plugins: {_plugin_save_to}")

    # Load each plugin,
    for plugin in _plugin_save_to:
        try:
            save_cls = plugin.load()
        except (ModuleNotFoundError, AttributeError) as e:
            debug_literal(
                f"plugins.load_from_functions: got error loading {plugin.name}: {str(e)}"
            )
            continue

        # get 'priority' if it is available
        priority = getattr(save_cls, "priority", DEFAULT_SAVE_TO_PRIORITY)

        # retrieve name (which is specified by plugin?)
        name = plugin.name
        debug_literal(f"plugins.save_to_functions: got '{name}', priority={priority}")
        yield priority, save_cls


class CommandLinePlugin:
    """
    Provide some minimal common CLI functionality - -q and -d.

    Subclasses should call super().__init__(parser) and super().main(args).
    """

    command = None
    description = None

    def __init__(self, parser):
        parser.add_argument(
            "-q", "--quiet", action="store_true", help="suppress non-error output"
        )
        parser.add_argument(
            "-d", "--debug", action="store_true", help="provide debugging output"
        )

    def main(self, args):
        set_quiet(args.quiet, args.debug)


def get_cli_script_plugins():
    global _plugin_cli_once

    x = []
    for plugin in _plugin_cli:
        name = plugin.name
        mod = plugin.module
        try:
            script_cls = plugin.load()
        except (ModuleNotFoundError, AttributeError):
            if _plugin_cli_once is False:
                error(
                    f"ERROR: cannot find or load module for cli_script plugin '{name}'"
                )
            continue

        command = getattr(script_cls, "command", None)
        if command is None:
            # print error message only once...
            if _plugin_cli_once is False:
                error(
                    f"ERROR: no command provided by cli_script plugin '{name}' from {mod}; skipping"
                )
        else:
            x.append(plugin)

    _plugin_cli_once = True
    return x


def get_cli_scripts_descriptions():
    "Build the descriptions for command-line plugins."
    for plugin in get_cli_script_plugins():
        name = plugin.name
        script_cls = plugin.load()

        command = getattr(script_cls, "command")
        description = getattr(script_cls, "description", "")
        if description:
            description = description.splitlines()[0]
        if not description:
            description = f"(no description provided by plugin '{name}')"

        yield f"sourmash scripts {command:16s} - {description}"


def add_cli_scripts(parser):
    "Configure parsing for command-line plugins."
    d = {}

    for plugin in get_cli_script_plugins():
        name = plugin.name
        script_cls = plugin.load()

        usage = getattr(script_cls, "usage", None)
        description = getattr(script_cls, "description", None)
        epilog = getattr(script_cls, "epilog", None)
        formatter_class = getattr(script_cls, "formatter_class", argparse.HelpFormatter)

        subparser = parser.add_parser(
            script_cls.command,
            usage=usage,
            description=description,
            epilog=epilog,
            formatter_class=formatter_class,
        )
        debug_literal(
            f"cls_script plugin '{name}' adding command '{script_cls.command}'"
        )
        obj = script_cls(subparser)
        d[script_cls.command] = obj

    return d


def list_all_plugins():
    plugins = itertools.chain(_plugin_load_from, _plugin_save_to, _plugin_cli)
    plugins = list(plugins)

    if not plugins:
        notify("\n(no plugins detected)\n")

    notify("")
    notify("the following plugins are installed:")
    notify("")
    notify(
        f"{'plugin type':<20s} {'from python module':<30s} {'v':<5s} {'entry point name':<20s}"
    )
    notify(f"{'-' * 20} {'-' * 30} {'-' * 5} {'-' * 20}")

    for plugin in plugins:
        name = plugin.name
        mod = plugin.module
        version = plugin.dist.version
        group = plugin.group

        notify(f"{group:<20s} {mod:<30s} {version:<5s} {name:<20s}")



import pytest
import sourmash_tst_utils as utils
from sourmash_tst_utils import SourmashCommandFailed


def test_failed_sourmash_exception(runtmp):
    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash("")



import gzip
from tempfile import NamedTemporaryFile

from screed.fasta import fasta_iter
import pytest

from sourmash.hll import HLL

import sourmash_tst_utils as utils

K = 21  # size of kmer
ERR_RATE = 0.01
N_UNIQUE = 3356
TRANSLATE = {"A": "T", "C": "G", "T": "A", "G": "C"}


def test_hll_add_python():
    # test python code to count unique kmers using HyperLogLog.
    # use the lower level add() method, which accepts anything,
    # and compare to an exact count using collections.Counter

    filename = utils.get_test_data("ecoli.genes.fna")
    hll = HLL(ERR_RATE, K)
    counter = set()

    with open(filename) as f:
        for n, record in enumerate(fasta_iter(f)):
            sequence = record["sequence"]
            seq_len = len(sequence)
            for n in range(0, seq_len + 1 - K):
                kmer = sequence[n : n + K]
                rc = "".join(TRANSLATE[c] for c in kmer[::-1])

                hll.add(kmer)

                if rc in counter:
                    kmer = rc
                counter.update([kmer])

    n_unique = len(counter)

    assert n_unique == N_UNIQUE
    assert abs(1 - float(hll.cardinality()) / N_UNIQUE) < ERR_RATE


def test_hll_consume_string():
    # test rust code to count unique kmers using HyperLogLog,
    # using screed to feed each read to the counter.

    filename = utils.get_test_data("ecoli.genes.fna")
    hll = HLL(ERR_RATE, K)
    n = 0
    with open(filename) as f:
        for n, record in enumerate(fasta_iter(f), 1):
            hll.add_sequence(record["sequence"])

    assert abs(1 - float(len(hll)) / N_UNIQUE) < ERR_RATE


def test_hll_similarity_containment():
    N_UNIQUE_H1 = 500741
    N_UNIQUE_H2 = 995845

    SIMILARITY = 0.502783
    CONTAINMENT_H1 = 1.0
    CONTAINMENT_H2 = 0.502783

    INTERSECTION = 500838

    hll1 = HLL(ERR_RATE, K)
    hll2 = HLL(ERR_RATE, K)
    hllu = HLL(ERR_RATE, K)

    filename = utils.get_test_data("genome-s10.fa.gz")
    with gzip.GzipFile(filename) as f:
        for n, record in enumerate(fasta_iter(f)):
            sequence = record["sequence"]
            seq_len = len(sequence)
            for n in range(0, seq_len + 1 - K):
                kmer = sequence[n : n + K]
                hll1.add(kmer)
                hllu.add(kmer)

    filename = utils.get_test_data("genome-s10+s11.fa.gz")
    with gzip.GzipFile(filename) as f:
        for n, record in enumerate(fasta_iter(f)):
            sequence = record["sequence"]
            seq_len = len(sequence)
            for n in range(0, seq_len + 1 - K):
                kmer = sequence[n : n + K]
                hll2.add(kmer)
                hllu.add(kmer)

    assert abs(1 - float(hll1.cardinality()) / N_UNIQUE_H1) < ERR_RATE
    assert abs(1 - float(hll2.cardinality()) / N_UNIQUE_H2) < ERR_RATE

    assert abs(1 - float(hll1.similarity(hll2)) / SIMILARITY) < ERR_RATE

    assert abs(1 - float(hll1.containment(hll2)) / CONTAINMENT_H1) < ERR_RATE
    assert abs(1 - float(hll2.containment(hll1)) / CONTAINMENT_H2) < ERR_RATE

    assert abs(1 - float(hll1.intersection(hll2)) / INTERSECTION) < ERR_RATE

    """
    hll1.merge(hll2)

    assert abs(1 - float(hllu.similarity(hll1))) < ERR_RATE

    assert abs(1 - float(hllu.containment(hll1))) < ERR_RATE
    assert abs(1 - float(hllu.containment(hll2))) < ERR_RATE

    assert abs(1 - float(hll1.intersection(hllu)) / N_UNIQUE_U) < ERR_RATE
    """


def test_hll_save_load():
    filename = utils.get_test_data("ecoli.genes.fna")
    hll = HLL(ERR_RATE, K)
    n = 0
    with open(filename) as f:
        for n, record in enumerate(fasta_iter(f), 1):
            hll.add_sequence(record["sequence"])

    assert abs(1 - float(len(hll)) / N_UNIQUE) < ERR_RATE

    with NamedTemporaryFile() as f:
        hll.save(f.name)

        new_hll = HLL.load(f.name)

    assert len(hll) == len(new_hll)



"""rename signature"""

usage = """

### `sourmash signature rename` - rename a signature

Rename the display name for one or more signatures - this is the name
output for matches in `compare`, `search`, `gather`, etc.

For example,

sourmash signature rename file1.sig "new name" -o renamed.sig

will place a renamed copy of the hashes in `file1.sig` in the file
`renamed.sig`. If you provide multiple signatures, all will be renamed
to the same name.

"""

from sourmash.cli.utils import (
    add_moltype_args,
    add_ksize_arg,
    add_picklist_args,
    add_pattern_args,
)


def subparser(subparsers):
    subparser = subparsers.add_parser("rename", description=__doc__, usage=usage)
    subparser.add_argument("signatures", nargs="*")
    subparser.add_argument("name")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-d", "--debug", action="store_true", help="print debugging output"
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        help="output renamed signature to this file (default stdout)",
        default="-",
    )
    subparser.add_argument(
        "-f", "--force", action="store_true", help="try to load all files as signatures"
    )
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_pattern_args(subparser)
    add_picklist_args(subparser)


def main(args):
    import sourmash

    return sourmash.sig.__main__.rename(args)



"""split signature files"""

usage = """

### `sourmash signature split` - split signatures into individual files

Split each signature in the input file(s) into individual files, with
standardized names.

For example,

sourmash signature split tests/test-data/2.fa.sig

will create 3 files,

`f372e478.k=21.scaled=1000.DNA.dup=0.2.fa.sig`,
`f3a90d4e.k=31.scaled=1000.DNA.dup=0.2.fa.sig`, and
`43f3b48e.k=51.scaled=1000.DNA.dup=0.2.fa.sig`, representing the three
different DNA signatures at different ksizes created from the input file
`2.fa`.

The format of the names of the output files is standardized and stable
for major versions of sourmash: currently, they are period-separated
with fields:

* `md5sum` - a unique hash value based on the contents of the signature.
* `k=<ksize>` - k-mer size.
* `scaled=<scaled>` or `num=<num>` - scaled or num value for MinHash.
* `<moltype>` - the molecule type (DNA, protein, dayhoff, or hp)
* `dup=<n>` - a non-negative integer that prevents duplicate signatures from colliding.
* `basename` - basename of first input file used to create signature; if none provided, or stdin, this is `none`.

If `--outdir` is specified, all of the signatures are placed in outdir.

Note: `split` only saves files in the JSON `.sig` format.

"""

from sourmash.cli.utils import add_moltype_args, add_ksize_arg, add_picklist_args


def subparser(subparsers):
    subparser = subparsers.add_parser("split", description=__doc__, usage=usage)
    subparser.add_argument("signatures", nargs="*")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "--output-dir",
        "--outdir",
        help="output signatures to this directory",
    )
    subparser.add_argument(
        "-f", "--force", action="store_true", help="try to load all files as signatures"
    )
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    subparser.add_argument(
        "-E",
        "--extension",
        type=str,
        default=".sig",
        help="write files with this extension ('.sig' by default)",
    )
    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_picklist_args(subparser)


def main(args):
    import sourmash

    return sourmash.sig.__main__.split(args)



import sourmash_tst_utils as utils


@utils.in_tempdir
def test_bug_803(c):
    # can we do a 'sourmash search' on an LCA database and a query with abundance?
    query = utils.get_test_data("track_abund/47.fa.sig")
    lca_db = utils.get_test_data("lca/47+63.lca.json")

    c.run_sourmash("search", query, lca_db, "--ignore-abundance")
    print(c)
    assert "NC_009665.1 Shewanella baltica OS185, complete genome" in str(c)



"""
Utility functions for lowest-common-ancestor analysis tools.
"""

from os.path import exists
from collections import namedtuple, defaultdict, Counter

from .lca_db import LCA_Database, load_single_database, load_databases


__all__ = [
    "taxlist",
    "zip_lineage",
    "build_tree",
    "find_lca",
    "load_single_database",
    "load_databases",
    "gather_assignments",
    "count_lca_for_assignments",
    "LineagePair",
    "display_lineage",
    "make_lineage",
    "pop_to_rank",
    "is_lineage_match",
]

try:  # py2/py3 compat
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

from sourmash.logging import notify, error, debug

# type to store an element in a taxonomic lineage
LineagePair = namedtuple("LineagePair", ["rank", "name"])


def check_files_exist(*files):
    ret = True
    not_found = []
    for f in files:
        if not exists(f):
            not_found.append(f)
            ret = False

    if len(not_found):
        error(
            "Error! Could not find the following files."
            " Make sure the file paths are specified correctly.\n{}".format(
                "\n".join(not_found)
            )
        )

    return ret


# ordered list of taxonomic ranks
def taxlist(include_strain=True):
    """
    Provide an ordered list of taxonomic ranks.
    """
    yield from [
        "superkingdom",
        "phylum",
        "class",
        "order",
        "family",
        "genus",
        "species",
    ]
    if include_strain:
        yield "strain"


# produce an ordered list of tax names from lineage
def zip_lineage(lineage, include_strain=True, truncate_empty=False):
    """
    Given an iterable of LineagePair objects, return list of lineage names.

    This utility function handles species/strain and empty lineage entries
    gracefully.

    >>> x = [ LineagePair('superkingdom', 'a'), LineagePair('phylum', 'b') ]
    >>> zip_lineage(x)
    ['a', 'b', '', '', '', '', '', '']

    >>> x = [ LineagePair('superkingdom', 'a'), LineagePair(None, ''), LineagePair('class', 'c') ]
    >>> zip_lineage(x)
    ['a', '', 'c', '', '', '', '', '']
    """

    empty = LineagePair(None, "")

    pairs = zip_longest(
        taxlist(include_strain=include_strain), lineage, fillvalue=empty
    )
    pairs = list(pairs)

    # eliminate empty if so requested
    if truncate_empty:
        last_lineage_tup = pairs[-1][1]
        while pairs and last_lineage_tup == empty:
            pairs.pop(-1)
            if pairs:
                last_lineage_tup = pairs[-1][1]

    row = []
    for taxrank, lineage_tup in pairs:
        # validate non-empty tax, e.g. superkingdom/phylum/class in order.
        if lineage_tup != empty and lineage_tup.rank != taxrank:
            raise ValueError(
                f"incomplete lineage at {taxrank} - is {lineage_tup.rank} instead"
            )

        row.append(lineage_tup.name)
    return row


def display_lineage(lineage, include_strain=True, truncate_empty=True):
    return ";".join(
        zip_lineage(
            lineage, include_strain=include_strain, truncate_empty=truncate_empty
        )
    )


# filter function toreplace blank/na/null with 'unassigned'
def filter_null(x):
    return (
        "unassigned" if x is None or x.strip() in ("[Blank]", "na", "null", "") else x
    )


null_names = set(["[Blank]", "na", "null"])


def build_tree(assignments, initial=None):
    """
    Builds a tree of dictionaries from lists of LineagePair objects
    in 'assignments'.  This tree can then be used to find lowest common
    ancestor agreements/confusion.
    """
    if initial is None:
        tree = {}
    else:
        tree = initial

    if not assignments:
        raise ValueError("empty assignment passed to build_tree")

    for assignment in assignments:
        node = tree

        for lineage_tup in assignment:
            if lineage_tup.name:
                child = node.get(lineage_tup, {})
                node[lineage_tup] = child

                # shift -> down in tree
                node = child

    return tree


def find_lca(tree):
    """
    Given a tree produced by 'find_tree', find the first node with multiple
    children, OR the only leaf in the tree.  Return (lineage_tup, reason),
    where 'reason' is the number of children of the returned node, i.e.
    0 if it's a leaf and > 1 if it's an internal node.
    """

    node = tree
    lineage = []
    while 1:
        if len(node) == 1:  # descend to only child; track path
            lineage_tup = next(iter(node.keys()))
            lineage.append(lineage_tup)
            node = node[lineage_tup]
        elif len(node) == 0:  # at leaf; end
            return tuple(lineage), 0
        else:  # len(node) > 1 => confusion!!
            return tuple(lineage), len(node)


def gather_assignments(hashvals, dblist):
    """
    Gather assignments from across all the databases for all the hashvals.

    Ignores counts of the hashvals.
    """
    assignments = defaultdict(set)
    for hashval in hashvals:
        for lca_db in dblist:
            lineages = lca_db.get_lineage_assignments(hashval)
            if lineages:
                assignments[hashval].update(lineages)

    return assignments


def count_lca_for_assignments(assignments, hashval_counts=None):
    """
    For each hashval, count the LCA across its assignments.

    If hashval_counts is not None, it must be a dictionary that maps
    { hashval: hashval_count }; this is then used to weight the counts.
    """
    counts = Counter()
    for hashval in assignments:
        # for each list of tuple_info [(rank, name), ...] build
        # a tree that lets us discover lowest-common-ancestor.
        lineages = assignments[hashval]
        tree = build_tree(lineages)

        # now find either a leaf or the first node with multiple
        # children; that's our lowest-common-ancestor node.
        lca, reason = find_lca(tree)

        if hashval_counts:
            counts[lca] += hashval_counts[hashval]
        else:
            counts[lca] += 1

    return counts


def is_lineage_match(lin_a, lin_b, rank):
    """
    check to see if two lineages are a match down to given rank.
    """
    for a, b in zip(lin_a, lin_b):
        assert a.rank == b.rank
        if a.rank == rank:
            if a == b:
                return 1
        if a != b:
            return 0

    return 0


def pop_to_rank(lin, rank):
    "Remove lineage tuples from given lineage `lin` until `rank` is reached."
    lin = list(lin)

    txl = taxlist()
    before_rank = []
    for txl_rank in txl:
        if txl_rank != rank:
            before_rank.append(txl_rank)
        else:
            break

    # are we already above rank?
    if lin and lin[-1].rank in before_rank:
        return tuple(lin)

    while lin and lin[-1].rank != rank:
        lin.pop()

    return tuple(lin)


def make_lineage(lineage):
    "Turn a ; or ,-separated set of lineages into a tuple of LineagePair objs."
    from sourmash.tax.tax_utils import LineagePair

    lin = lineage.split(";")
    if len(lin) == 1:
        lin = lineage.split(",")
    lin = [LineagePair(rank, n) for (rank, n) in zip(taxlist(), lin)]
    lin = tuple(lin)

    return lin



import random
from tempfile import NamedTemporaryFile

from sourmash.sbt_storage import ZipStorage
from sourmash.minhash import MinHash

RANDOM_SEQ_SIZE = 3000
RANDOM_SEQ_NUMBER = 300

MINHASH_NUM = 500
MINHASH_K = 21

GET_MINS_RANGE = 500
ADD_HASH_RANGE = 10_000
ADD_MANY_RANGE = 1000
SIMILARITY_TIMES = 500
COUNT_COMMON_TIMES = 500
MERGE_TIMES = 500
COPY_TIMES = 500
CONCAT_TIMES = 500
SET_ABUNDANCES_RANGE = 500
ZIP_STORAGE_WRITE = 100_000
ZIP_STORAGE_LOAD = 20


def load_sequences():
    sequences = []
    for i in range(10):
        random_seq = random.sample(
            "A,C,G,T".split(",") * RANDOM_SEQ_SIZE, RANDOM_SEQ_NUMBER
        )
        sequences.append("".join(random_seq))
    return sequences


class TimeMinHashSuite:
    def setup(self):
        self.mh = MinHash(MINHASH_NUM, MINHASH_K, track_abundance=False)
        self.protein_mh = MinHash(
            MINHASH_NUM, MINHASH_K, is_protein=True, track_abundance=False
        )
        self.sequences = load_sequences()

        self.populated_mh = MinHash(MINHASH_NUM, MINHASH_K, track_abundance=False)
        for seq in self.sequences:
            self.populated_mh.add_sequence(seq)

    def time_add_sequence(self):
        mh = self.mh
        sequences = self.sequences
        for seq in sequences:
            mh.add_sequence(seq)

    def time_add_protein(self):
        mh = self.protein_mh
        sequences = self.sequences
        for seq in sequences:
            mh.add_protein(seq)

    def time_get_mins(self):
        mh = self.populated_mh
        for i in range(GET_MINS_RANGE):
            mh.get_mins()

    def time_add_hash(self):
        mh = self.mh
        for i in range(ADD_HASH_RANGE):
            mh.add_hash(i)

    def time_add_many(self):
        mh = self.mh
        mh.add_many(list(range(ADD_MANY_RANGE)))

    def time_similarity(self):
        mh = self.mh
        other_mh = self.populated_mh
        for i in range(SIMILARITY_TIMES):
            mh.similarity(other_mh)

    def time_count_common(self):
        mh = self.mh
        other_mh = self.populated_mh
        for i in range(COUNT_COMMON_TIMES):
            mh.count_common(other_mh)

    def time_merge(self):
        mh = self.mh
        other_mh = self.populated_mh
        for i in range(MERGE_TIMES):
            mh.merge(other_mh)

    def time_copy(self):
        mh = self.populated_mh
        for i in range(COPY_TIMES):
            mh.__copy__()

    def time_concat(self):
        mh = self.mh
        other_mh = self.populated_mh
        for i in range(CONCAT_TIMES):
            mh += other_mh


class PeakmemMinHashSuite:
    def setup(self):
        self.mh = MinHash(MINHASH_NUM, MINHASH_K, track_abundance=True)
        self.protein_mh = MinHash(
            MINHASH_NUM, MINHASH_K, is_protein=True, track_abundance=True
        )
        self.sequences = load_sequences()

    def peakmem_add_sequence(self):
        mh = self.mh
        sequences = self.sequences
        for seq in sequences:
            mh.add_sequence(seq)

    def peakmem_add_protein(self):
        mh = self.protein_mh
        sequences = self.sequences
        for seq in sequences:
            mh.add_protein(seq)

    def peakmem_add_hash(self):
        mh = self.mh
        for i in range(ADD_HASH_RANGE):
            mh.add_hash(i)

    def peakmem_add_many(self):
        mh = self.mh
        mh.add_many(list(range(ADD_MANY_RANGE)))


####################


class TimeMinAbundanceSuite(TimeMinHashSuite):
    def setup(self):
        TimeMinHashSuite.setup(self)
        self.mh = MinHash(MINHASH_NUM, MINHASH_K, track_abundance=True)

        self.populated_mh = MinHash(MINHASH_NUM, MINHASH_K, track_abundance=True)
        for seq in self.sequences:
            self.populated_mh.add_sequence(seq)

    def time_get_mins_abundance(self):
        mh = self.populated_mh
        for i in range(GET_MINS_RANGE):
            mh.get_mins(with_abundance=True)

    def time_set_abundances(self):
        mh = self.mh
        mins = self.populated_mh.get_mins(with_abundance=True)
        for i in range(SET_ABUNDANCES_RANGE):
            mh.set_abundances(mins)

    def time_set_abundances_noclear(self):
        mh = self.mh
        mins = self.populated_mh.get_mins(with_abundance=True)
        for i in range(SET_ABUNDANCES_RANGE):
            mh.set_abundances(mins, clear=False)


class PeakmemMinAbundanceSuite(PeakmemMinHashSuite):
    def setup(self):
        PeakmemMinHashSuite.setup(self)
        self.mh = MinHash(MINHASH_NUM, MINHASH_K, track_abundance=True)


####################


class TimeZipStorageSuite:
    def setup(self):
        import zipfile

        self.zipfile = NamedTemporaryFile()

        with zipfile.ZipFile(
            self.zipfile, mode="w", compression=zipfile.ZIP_STORED
        ) as storage:
            for i in range(ZIP_STORAGE_WRITE):
                # just so we have lots of entries
                storage.writestr(str(i), b"0")
            # one big-ish entry
            storage.writestr("sig1", b"9" * 1_000_000)

    def time_load_from_zipstorage(self):
        with ZipStorage(self.zipfile.name) as storage:
            for i in range(ZIP_STORAGE_LOAD):
                storage.load("sig1")

    def time_load_small_from_zipstorage(self):
        with ZipStorage(self.zipfile.name) as storage:
            for i in range(ZIP_STORAGE_LOAD):
                storage.load("99999")

    def teardown(self):
        self.zipfile.close()


class PeakmemZipStorageSuite:
    def setup(self):
        import zipfile

        self.zipfile = NamedTemporaryFile()

        with zipfile.ZipFile(
            self.zipfile, mode="w", compression=zipfile.ZIP_STORED
        ) as storage:
            for i in range(ZIP_STORAGE_WRITE):
                # just so we have lots of entries
                storage.writestr(str(i), b"0")
            # one big-ish entry
            storage.writestr("sig1", b"9" * 1_000_000)

    def peakmem_load_from_zipstorage(self):
        with ZipStorage(self.zipfile.name) as storage:
            for i in range(ZIP_STORAGE_LOAD):
                storage.load("sig1")

    def peakmem_load_small_from_zipstorage(self):
        with ZipStorage(self.zipfile.name) as storage:
            for i in range(ZIP_STORAGE_LOAD):
                storage.load("99999")

    def teardown(self):
        self.zipfile.close()



import pytest

from hypothesis import given, example
import hypothesis.strategies as st

from sourmash import MinHash
from sourmash.minhash import _get_max_hash_for_scaled


@given(
    st.lists(st.integers(min_value=0, max_value=2**64 - 1), min_size=10, max_size=1000),
    st.lists(st.integers(min_value=0, max_value=2**64 - 1), min_size=10, max_size=1000),
    st.integers(min_value=10, max_value=1000),
)
@example([1, 2], [3, 4], 2)
def test_set_abundance_num_hypothesis(hashes, abundances, sketch_size):
    a = MinHash(sketch_size, 10, track_abundance=True)
    oracle = dict(zip(hashes, abundances))

    a.set_abundances(oracle)

    mins = a.hashes
    size = min(sum(1 for v in oracle.values() if v > 0), sketch_size)
    assert len(mins) == size

    for k, v in mins.items():
        assert oracle[k] == v


@given(
    st.lists(st.integers(min_value=0, max_value=2**64 - 1), min_size=10, max_size=1000),
    st.lists(st.integers(min_value=0, max_value=2**64 - 1), min_size=10, max_size=1000),
    st.integers(min_value=1000, max_value=10000),
)
@example([0], [0], 1000)
def test_set_abundance_scaled_hypothesis(hashes, abundances, scaled):
    a = MinHash(0, 10, track_abundance=True, scaled=scaled)
    oracle = dict(zip(hashes, abundances))

    a.set_abundances(oracle)

    max_hash = _get_max_hash_for_scaled(scaled)
    below_max_hash = sum(1 for (k, v) in oracle.items() if k <= max_hash and v > 0)

    mins = a.hashes
    assert len(mins) == below_max_hash

    for k, v in mins.items():
        assert oracle[k] == v
        assert k <= max_hash
        assert v > 0



"'sourmash categorize' - query an SBT for bes match, with many signatures."

from sourmash.cli.utils import add_ksize_arg, add_moltype_args


def subparser(subparsers):
    subparser = subparsers.add_parser("categorize")
    subparser.add_argument(
        "database", help="location of signature collection/database to load"
    )
    subparser.add_argument(
        "queries", nargs="+", help="locations of signatures to categorize"
    )
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    add_ksize_arg(subparser)
    subparser.add_argument(
        "--threshold",
        default=0.08,
        type=float,
        help="minimum threshold for reporting matches; default=0.08",
    )
    subparser.add_argument(
        "--ignore-abundance",
        action="store_true",
        help="do NOT use k-mer abundances if present",
    )
    add_moltype_args(subparser)

    # TODO: help messages in these
    subparser.add_argument("--csv", help="output summary CSV to this file")
    subparser.add_argument("--load-csv", default=None)


def main(args):
    import sourmash

    return sourmash.commands.categorize(args)
